package displayPack;

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Graphics;
import java.util.Vector;

import javax.swing.JScrollBar;

import modelPack.FSNode;
import modelPack.LanguageModel;
import modelPack.Slot;
import modelPack.SlotValue;
import utilPack.Util;

// I have tried extending Canvas (AWT, therefore mixing lightweight with Swing), JPanel and JComponent

public class DrawArea extends Canvas{
	
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	private VFrame vf;
	
	//----------------------------------------------------------------------------------
	//                            scrollbars
	//----------------------------------------------------------------------------------
	
		
	// value of the horizontal scrollbar in the main frame - in a range 0..90
	private int horizontalScrollValue = 0;
	public void setHorizontalScrollValue(int value) {horizontalScrollValue = value;}
	
	// move all images to the left by horizontalScrollScale pixels for each unit increase in the scroll value
	private int horizontalScrollScale = 50;
		
	// left movement of all images
	private int leftScrollShift() {return horizontalScrollScale*horizontalScrollValue;}
	

	/* This happens when the vertical scrollbar is adjusted,
	 * value of the vertical scrollbar in the main frame - in a range 0..90 */
	public void setVerticalScrollValue(int value) 
	{
		ignoreScrollBar = false;
		upDisplacement = verticalScrollScale*value;
	}
		
	/**
	 * set the vertical scroll position so that pixel at height y = upDisplacement
	 *  is at the top of the visible area
	 */
	private void adjustScrollBar()
	{
		verticalScrollBar.setValue(upDisplacement/verticalScrollScale);
		vf.validate();
	}

	// move all images up by verticalScrollScale pixels for each unit increase in the scroll value
	private int verticalScrollScale = 100;
	
	// the number of pixels to move the image up
	private int upDisplacement = 0;
	
	private JScrollBar verticalScrollBar;
	public void addVertScrollBar(JScrollBar verticalScrollBar) {this.verticalScrollBar = verticalScrollBar;}
	
	
	private boolean ignoreScrollBar = false;
	public void setIgnoreScrollBar(boolean ignore) {ignoreScrollBar = ignore;}
	
	

	
	//----------------------------------------------------------------------------------
	//                            
	//----------------------------------------------------------------------------------
	

	// slot definition to be drawn
	private Slot slotToDraw = null;
	public void setSlotToDraw(Slot s) 
	{
		slotToDraw = s;
	}
	public Slot getDrawnSlot() {return slotToDraw;}
	
	// rows of feature structures to be drawn
	private Vector<Vector<FSNode>> fssToDraw = new Vector<Vector<FSNode>>();
	public Vector<Vector<FSNode>> getAllDrawnFS() {return fssToDraw;}
	
	
	// whether or not to show nodes used in previous unifications
	private Boolean showUsedNodes = false;
	public void setShowUsedNodes(boolean show) {showUsedNodes = show;}
	public boolean showUsedNodes() {return showUsedNodes;}
	
	/**
	 * clear all rows and make an empty first row
	 */
	public void clearAllRows() 
	{
		fssToDraw = new Vector<Vector<FSNode>>();
		fssToDraw.add(new Vector<FSNode>());
	}
	
	/**
	 * clear the last row, if there is one;
	 * if not, create an empty  row
	 */
	public void clearLastRow() 
	{
		if (fssToDraw.size() > 0) fssToDraw.get(fssToDraw.size()-1).removeAllElements();
		else fssToDraw.add(new Vector<FSNode>());
	}
	
	/**
	 * @return the last row of feature structures, if there is one;
	 * otherwise, make one and return it
	 */
	public Vector<FSNode> getLastRow() 
	{
		if (fssToDraw.size() > 0) return fssToDraw.get(fssToDraw.size()-1);
		else
		{
			Vector<FSNode> lastRow = new Vector<FSNode>();
			fssToDraw.add(lastRow);
			return lastRow;
		}
	}

	/**
	 * add a new empty row of FS
	 */
	public void addNewRow() 
	{
		fssToDraw.add(new Vector<FSNode>());
	}
	
	/**
	 * 
	 * @param toDraw
	 */
	public void extendLastRow(FSNode toDraw)
	{
		getLastRow().add(toDraw);
		repaint();
	}
	
	
	
	public FSNode getFirstFS()
	{
		FSNode fs = null;
		if ((getLastRow() != null) && (getLastRow().size() > 0)) fs = getLastRow().get(0);
		return fs;
	}
	
	public FSNode getDrawnFS(int i)
	{
		FSNode drawn = null;
		
		if ((getLastRow() != null) && (getLastRow().size() > i))  drawn = getLastRow().get(i);
		
		return drawn;
	}
	
	// add a feature structure to be drawn, below the others
	public void addToDraw(FSNode f) 
	{
		if (getLastRow() != null) getLastRow().add(f);
		else Util.message("failure in addToDraw: no last row");
	}
		
	
	/**
	 * clear the last row and draw up to three feature structures on it, 
	 * possibly keeping the first
	 * 
	 * @param keepFirst
	 * @param first
	 * @param second
	 * @param third
	 */
	public void drawFS(boolean keepFirst, FSNode first, FSNode second, FSNode third)
	{
		FSNode zeroth = null;
		if ((getLastRow() != null) && (getLastRow().size() > 0) && keepFirst)  zeroth = getLastRow().get(0);
		
		clearLastRow();
		
		if (zeroth != null) addToDraw(zeroth);
		if (first != null) addToDraw(first);
		if (second != null) addToDraw(second);
		if (third != null) addToDraw(third);
		
		repaint();
	}
	
	/**
	 * clear everything and draw a slot definition
	 * @param s
	 */
	public void drawSlot(Slot s)
	{
		clearAllRows();
		slotToDraw = s;
		repaint();
	}
	
	/**
	 * clear everything  (eg on finishing editing the model)
	 */
	public void clear()
	{
		clearAllRows();
		slotToDraw = null;
		repaint();
	}
	
    
    // vertical gap between Feature structures
	public static int drawGapY = 20;
	
	// horizontal gap between feature structure
	public static int drawGapX = 60;
	
	// top and left of a slot tree being drawn
	public static int slotLeft = 60;
	public static int slotTop = 60;

    
    //--------------------------------------------------------------------------------------
    //                              Constructor
    //--------------------------------------------------------------------------------------
	
	public DrawArea(VFrame vf)
	{
		this.vf = vf;
		setVisible(true);
		
	    unsetMouseCoords();
	    detectMouseMotion dMouseMotion = new detectMouseMotion();
	    addMouseListener(dMouseMotion); // to detect entry and exit from the grid area

	}
	
	/**
	 * @return
	 */
	public int totalHeight(int rowNo)
	{
		int h = drawGapY;
		
		if (rowNo < fssToDraw.size())
			for (FSNode f:fssToDraw.get(rowNo))
			{
				// true = include the height of any caption below a tree diagram
				int thisH = 0;
				try{thisH = drawGapY + f.getDrawingHeight(true,showUsedNodes);}
				catch(Exception ex) {ex.printStackTrace();}
				if (thisH > h) h = thisH;
			}

		return h;
	}
	
	public int totalWidth(int rowNo)
	{
		int w = drawGapX;
		
		if (rowNo < fssToDraw.size())
			for (FSNode f:fssToDraw.get(rowNo))
				w = w + f.getDrawingWidth() + drawGapX;
		return w;		
	}
	

	/**
	 * 
	 */
	public void paint(Graphics g) 
	  {
		 			 
			 /* clear the clickable rectangles of Feature Structures or SlotValues, 
			  * which when drawn below, will populate the arrays of Rectangles used to detect mouse clicks in them*/
			 if (vf.getModel() != null) 
			 {
				 vf.getModel().clearNodeLocations();
				 vf.getModel().clearValueLocations();
				 vf.getModel().clearpmBoxLocations();
			 }
			    
			 setBackground(Color.WHITE);
			 
			 // draw a slot if it exists 
			 if (slotToDraw != null)
			 {
				 slotToDraw.draw(slotLeft, slotTop, g);
			 }
			 
			 /* compute the amount of up displacement of the image from all but the last row of feature structures,
			  * to get the last row on screen */
			 if (ignoreScrollBar)
			 {
				 upDisplacement = 0;
				 for (int i = 0; i < fssToDraw.size() - 1; i++) 
					 upDisplacement = upDisplacement + totalHeight(i);			    
			 }
			 int y = drawGapY - upDisplacement;
			 
			 // iterate over rows of feature structures			    
			 for (int i = 0; i < fssToDraw.size(); i++)			    
			 {
				 Vector<FSNode> row = fssToDraw.get(i);
				 int x0 = drawGapX - leftScrollShift(); // start of the row
				 int x = x0; // start of latest FS
				 
				 // if this is not the first row, draw a line under the previous row
				 if ( i > 0) g.drawLine(x0, y + 20 , x0 + totalWidth(i-1), y +20);

				 y = y + 40;
				 
				 // draw across the row
				 for (int j = 0; j < row.size(); j++) try
				 {
					// show/hide should always be enabled 
					vf.showHideButton.setEnabled(true);
					
					//draw the FS
					row.get(j).draw(x,y,g,showUsedNodes);
					 
					 // prepare to draw the next feature structure to the right of this one
					 x = x + row.get(j).getDrawingWidth() + drawGapX;
				 }
				 catch(Exception ex) {ex.printStackTrace();}
				 
				 // increment y for the next row
				 y = y + totalHeight(i);				 
			 }
			 
			 // set the scrollbar to the right place
			 adjustScrollBar();

	  }
	
//-------------------------------------------------------------------------------
//                     Mouse movements and presses             
//-------------------------------------------------------------------------------
	
	  private void unsetMouseCoords()
	  {
		  dragX = 0;
		  dragY = 0;
	  }

  
  class detectMouseMotion extends java.awt.event.MouseAdapter
  {
      public void mousePressed(java.awt.event.MouseEvent event)
      {
    	  LanguageModel model = vf.getModel();
    	  
    	  // record where the press happened
    	  dragX = event.getX();
    	  dragY = event.getY();
    	  
    	  // test if the location clicked is in a node on a drawn Feature structure 
    	  FSNode fs = model.getClickedNode(dragX, dragY);
    	  clickedNode = fs;
    	  if (fs != null) 
    	  {
    		  // write out the origin
    		  // fs.writePositionAndOrigin();
    		  
    		  // highlight origins of the clicked node in the two input FS
    		  highlightOrigins(fs);
    		  
    		  // re-enable the language edit menus
    		  vf.getEditLanguage().notifySelect(fs);
    	  }
    	  
    	  // test if the location clicked is in a node on a drawn Slot Value 
    	  SlotValue sv = model.getClickedValue(dragX, dragY);
    	  clickedValue = sv;
    	  if (sv != null) 
    	  {
    		  // re-enable the model edit menus
    		  vf.getEditModel().notifySelect(sv);
    	  }
    	  
       	  
    	  // test if the location clicked is in a [+-] box on a drawn Slot Value 
    	  SlotValue pm = model.getClickedpmBox(dragX, dragY);
    	  if ((pm != null) && (event.getClickCount() == 1))
    	  {
    		  // toggle the expanded state of the slot value, before redrawing
    		  pm.setExpanded(!pm.isExpanded());
    	  }

    	  repaint();
      }
  }
  
  /**
   * highlight nodes in the two inputs which this result node came from
   * @param fs
   */
  private void highlightOrigins(FSNode fs)
  {
	  // if both inputs and a result are shown
	  if (getLastRow().size() == 3)
	  {
		  int[] origin = fs.getOrigin();
		  if (origin != null)			  
		  {
			  getLastRow().get(0).setHighlightedPosition(origin[0]);
			  getLastRow().get(1).setHighlightedPosition(origin[1]);
		  }
	  }
  }

  
	
	// current coordinates of the mouse as it is dragged
	private int dragX, dragY;
	
	// latest drag position of the mouse
	public int getDragX() {return dragX;}
	public int getDragY() {return dragY;}
	
	// used by FeatureStructures when drawing, to detect if they have been clicked
	private FSNode clickedNode;
	public FSNode getJustClickedNode() {return clickedNode;}
	
	// used by SlotValues when drawing, to detect if they have been clicked
	private SlotValue clickedValue;
	public SlotValue getJustClickedValue() {return clickedValue;}
    	  

}
