package displayPack;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.KeyStroke;

import modelPack.FSNode;
import modelPack.LanguageModel;
import modelPack.LearningStage;
import modelPack.NodeType;
import modelPack.Slot;
import modelPack.SlotValue;
import modelPack.WordToLearn;
import utilPack.Util;

public class EditLanguage {
	
	/**
	 *  key variables defining the state of the editor
	 */
	
	// whether the FS currently drawn is being edited - and if so, its type
	private int editState = NONE;
	public int getEditState() {return editState;}
	public void setEditState(int newState) {editState = newState;}
	
	// true if the shown FS can be edited
	public boolean canEdit() {return (editState != NONE);}
	
	// type of the FS being edited - is it a word, a sentence , or a learning example?
	private int shownFSType = NONE;
	public int getShownFSType() {return shownFSType;}
	public void setShownFSType(int type) {shownFSType = type;}
	
	// editing or not
	private boolean editing = false;
	public boolean isEditing() {return editing;}

	// possible Feature Structure types
	public static int NONE = -1;
	public static int WORD = 1;
	public static int SENTENCE = 2;
	public static int LEARNINGEXAMPLE = 3;
	public static int WORDTOLEARN = 4;
	
	// Possible actions on Feature Structures
	public static int SHOW = 0;
	public static int EDIT = 1;
	public static int REMOVE = 2;

	// node being edited
	private FSNode nodeToEdit = null;

	// true if the shown FS has been changed since the last save
	protected boolean changedSinceSaved;
	
	// node most recently selected by mouse click
	private FSNode selectedNode = null;

	// node copied, if 'copy' has been selected
	private FSNode copiedNode = null;
	
	// node selected to be the start of a link
	private FSNode linkStartNode = null;
	
	// link id to be added to both ends
	private String linkId = null;
	
	private DrawArea drawArea;

	private VFrame vf;
	
	private LanguageModel model() {return vf.getModel();} 
	
	
	JMenu editMenu = new JMenu("Edit Language");
	JMenu editFSMenu = new JMenu("Edit");
		JMenu editSlotMenu = new JMenu("Slots");
		JMenu editOtherValueMenu = new JMenu("Other values");
		JMenu editWordMenu = new JMenu("Word");
		JMenu editSentenceMenu = new JMenu("Sentence");
		JMenu editWordToLearnMenu = new JMenu("Word to Learn");
		JMenu editLearningExampleMenu = new JMenu("Learning Example");
	JMenuItem addMenu = new JMenu("Add");
		JMenu addChildNodeMenu = new JMenu("Child Node");
		JMenu addLinkMenu = new JMenu("Link");
			JMenuItem addLinkStartItem = new JMenuItem("Start");
			JMenuItem addLinkEndItem = new JMenuItem("End");
		JMenuItem addWordItem = new JMenuItem("Word");
		JMenuItem addSentenceItem = new JMenuItem("Sentence");
		JMenu addLearningExampleMenu = new JMenu("Learning Example");
		JMenu addWordToLearnMenu = new JMenu("Word to Learn");
		JMenu addSoundMenu = new JMenu("Sounds");
			JMenuItem soundBelowItem = new JMenuItem("Below");
			JMenuItem soundAfterItem = new JMenuItem("After");
			JMenuItem soundBeforeItem = new JMenuItem("Before");
	JMenu removeMenu = new JMenu("Remove");
		JMenu removeSlotMenu = new JMenu("Slot");
		JMenuItem removeNodeItem = new JMenuItem("Node");
		JMenuItem removeLinkItem = new JMenuItem("Link");
		JMenu removeWordMenu = new JMenu("Word");
		JMenu removeSentenceMenu = new JMenu("Sentence");
		JMenu removeLearningExampleMenu = new JMenu("Learning Example");
		JMenu removeWordToLearnMenu = new JMenu("Word to Learn");
	JMenuItem copyItem = new JMenuItem("Copy");
	JMenu pasteMenu = new JMenu("Paste");
		JMenuItem pasteOnItem = new JMenuItem("On");
		JMenuItem pasteBelowItem = new JMenuItem("Below");
	JMenu moveMenu = new JMenu("Move Node");
		JMenuItem moveUpItem = new JMenuItem("Up");
		JMenuItem moveDownItem = new JMenuItem("Down");
	JMenuItem saveItem = new JMenuItem("Save");
	JMenuItem finishItem = new JMenuItem("Finish");


	  //------------------------------------------------------------------------------------------------
	  //                              constructor and startup
	  //------------------------------------------------------------------------------------------------
	
	
	public EditLanguage(VFrame vf, DrawArea drawArea)
	{
		this.vf = vf;
		this.drawArea = drawArea;
		editing = false;
	}
	

	  //------------------------------------------------------------------------------------------------
	  //                              setting up the edit menus
	  //------------------------------------------------------------------------------------------------
	/**
	 * 
	 */
	public void setUpEditMenu(JMenu editMenu)
	{
		this.editMenu = editMenu;
		
	    editMenu.add(editFSMenu);
	    	editFSMenu.add(editSlotMenu);
	    	editFSMenu.add(editOtherValueMenu);
	    	editFSMenu.add(editWordMenu);
	    	editFSMenu.add(editSentenceMenu);
	    	editFSMenu.add(editWordToLearnMenu);
	    	editFSMenu.add(editLearningExampleMenu);
	    editMenu.add(addMenu);
	    	addMenu.add(addChildNodeMenu);
	    	addMenu.add(addLinkMenu);
	    		addLinkMenu.add(addLinkStartItem);
	    		addLinkMenu.add(addLinkEndItem);
	    	addMenu.add(addWordItem);
	    	addMenu.add(addSentenceItem);
    		addMenu.add(addLearningExampleMenu);
    		addMenu.add(addWordToLearnMenu);
		    addMenu.add(addSoundMenu);
    			addSoundMenu.add(soundBelowItem);
    			addSoundMenu.add(soundBeforeItem);
    			addSoundMenu.add(soundAfterItem);
    	editMenu.add(removeMenu);
        	removeMenu.add(removeNodeItem);
        	removeMenu.add(removeLinkItem);
        	removeMenu.add(removeSlotMenu);
        	removeMenu.add(removeWordMenu);
        	removeMenu.add(removeSentenceMenu);
        	removeMenu.add(removeLearningExampleMenu);
        	removeMenu.add(removeWordToLearnMenu);
	    editMenu.add(copyItem);
	    copyItem.setMnemonic(KeyEvent.VK_C);
	    editMenu.add(pasteMenu);
	    	pasteMenu.add(pasteOnItem);
	    	pasteMenu.add(pasteBelowItem);
		editMenu.add(moveMenu);
			moveMenu.add(moveUpItem);
			moveMenu.add(moveDownItem);
	    editMenu.add(saveItem);
    	editMenu.add(finishItem);
	    
	    // many top level sub-menus and items are disabled until you start to edit a FS
	    addChildNodeMenu.setEnabled(false);
	    editSlotMenu.setEnabled(false);
	    copyItem.setEnabled(false);
	    pasteMenu.setEnabled(false);
	    addSoundMenu.setEnabled(false);
	    removeMenu.setEnabled(false);
	    saveItem.setEnabled(false);
    	
    	finishItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	doFinish();
	        }
	    });		
    	
    	addWordItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	doNewFS(WORD);
	        }
	    });
    	
    	
    	addSentenceItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	doNewFS(SENTENCE);
	        }
	    });
    	
    	// design pattern to get a hot key 'ctrl C'  - it works
    	Action copyAction = new AbstractAction("Copy") {
 			private static final long serialVersionUID = 1L;
			public void actionPerformed(ActionEvent e) {
    			doCopy();
    		}
    	}; 
    	copyAction.putValue(Action.ACCELERATOR_KEY,KeyStroke.getKeyStroke(KeyEvent.VK_C, KeyEvent.CTRL_DOWN_MASK));    	
    	copyItem.setAction(copyAction);
    	
    	addLinkStartItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	doAddLinkStart();
	        }
	    });
    	
    	addLinkEndItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	doAddLinkEnd();
	        }
	    });
    	
    	removeLinkItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	doRemoveLink();
	        }
	    });
    	
    	pasteOnItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	doPasteOn();
	        }
	    });
    	
    	pasteBelowItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	doPasteBelow();
	        }
	    });

       	Action moveUpAction = new AbstractAction("Up") {
     			private static final long serialVersionUID = 1L;
    			public void actionPerformed(ActionEvent e) {
        			doMove(true);
        		}
        	}; 
        moveUpAction.putValue(Action.ACCELERATOR_KEY,KeyStroke.getKeyStroke(KeyEvent.VK_U, KeyEvent.CTRL_DOWN_MASK));    	
        moveUpItem.setAction(moveUpAction);
       	
       	Action moveDownAction = new AbstractAction("Down") {
     			private static final long serialVersionUID = 1L;
    			public void actionPerformed(ActionEvent e) {
        			doMove(false);
        		}
        	}; 
        moveDownAction.putValue(Action.ACCELERATOR_KEY,KeyStroke.getKeyStroke(KeyEvent.VK_D, KeyEvent.CTRL_DOWN_MASK));    	
        moveDownItem.setAction(moveDownAction);
    	
    	removeNodeItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	doRemoveNode();
	        }
	    });
     	
    	soundBelowItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	doSound(0);
	        }
	    });
    	
    	soundBeforeItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	doSound(1);
	        }
	    });
    	
    	soundAfterItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	doSound(2);
	        }
	    });

    	saveItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	doSave();
	        }
	    });
	}
	
	/**
	 * called when the user selects any node on any feature structure
	 * @param fs
	 */
	public void notifySelect(FSNode fs)	
	{
		// always happens (any selected node can be copied)
		selectedNode = fs;
		String type = fs.getType();

		Vector<FSNode> drawn = drawArea.getLastRow();
		String dd = "\ndrawn FS: ";
		for (FSNode d:drawn) dd = dd + d.getName() + "; ";
		
		// find the ultimate ancestor of the selected node
		FSNode selectedRoot = fs.getRootAncestor();
		String name = selectedRoot.getName();
		// Util.message("Selected node type: " + type + "; FS name: " + name);
		
		// the selected node can be edited or pasted to; it is on an FS being edited
		boolean selectedEditable = ((canEdit()) && (name.equals(drawn.get(0).getName())));
		if (selectedEditable) nodeToEdit = fs;
		else nodeToEdit = null;
		
		// you can copy a node from any FS (whether or not it is the one being edited)
		copyItem.setEnabled(true);
		
		// refresh all menus for the selected node
		refreshEditLanguageMenus(fs,copiedNode,selectedEditable);
	}
	
	/**
	 * enable or disable edit menu items, and create those that depend on data 
	 */
	
	/**
	 * 
	 * @param editableNode node currently being edited
	 * @param copNode node copied - ready to paste
	 * @param editable
	 */
	public void refreshEditLanguageMenus(FSNode editableNode, FSNode copNode , boolean editable)
	{
		// the edit menu structure is not sent to the client
		MenuElement dummy = new MenuElement("dummy");
	    
	    // menus to choose a Feature Structure to edit or remove
	    if (model() != null)
	    {
		    VFrame.addFeatureStructureItems(model().getWords(), vf.wordMenu, vf.getAct(), WORD, SHOW,dummy);
		    VFrame.addFeatureStructureItems(model().getSentences(), vf.sentenceMenu, vf.getAct(), SENTENCE, SHOW,dummy);

		    VFrame.addFeatureStructureItems(model().getWords(), editWordMenu, vf.getAct(), WORD, EDIT,dummy);
		    VFrame.addFeatureStructureItems(model().getSentences(), editSentenceMenu, vf.getAct(), SENTENCE, EDIT,dummy);

		    VFrame.addFeatureStructureItems(model().getWords(), removeWordMenu, vf.getAct(), WORD, REMOVE,dummy);
		    VFrame.addFeatureStructureItems(model().getSentences(), removeSentenceMenu, vf.getAct(), SENTENCE, REMOVE,dummy);

		    // sort out all menus for words to learn or learning examples
		    makeExampleEditMenus();
	    }		
		
	    // many sub-menus are disabled until you start to edit a FS
	    copyItem.setEnabled(editableNode != null); 
	    pasteMenu.setEnabled(false);
	    removeMenu.setEnabled(canEdit());
	    removeNodeItem.setEnabled(false); // altered if a node is selected
		
		editSlotMenu.removeAll();
	    editSlotMenu.setEnabled(canEdit());
		addChildNodeMenu.removeAll();
	    addChildNodeMenu.setEnabled(canEdit());
	    
	    // if there is a selected node, and it can be edited
		if ((editableNode != null) && (editable))
		{
			String type = editableNode.getType();
			// to remove a script node, remove the whole FS
			removeNodeItem.setEnabled(!type.equals("script"));
			
			makeNodeChangeMenus(type);
			makeNodeAddMenus(type);
						
			makeRemoveSlotMenu(editableNode);
			
			makeNodeValueMenuItems();
			
			enableMoveItems(editableNode);
			
			enableLinkItems(editableNode);
			
			enableSoundItems(editableNode);
		}	
		
		finishItem.setEnabled(editing);
		
		addWordItem.setEnabled(true);
		addSentenceItem.setEnabled(true);
		
		// if the selected node is editable, and there is a copied node, enable the paste menu items
		if ((copNode != null) && (editableNode != null) && (editable))
		{
			String cType = copNode.getType();
			String eType = editableNode.getType();
			pasteBelowItem.setEnabled(model().getNodeType(eType).isAllowedChild(cType));
			pasteOnItem.setEnabled ((cType.equals(eType)) && (!editableNode.hasChildren()));
		}
		else
		{
			pasteBelowItem.setEnabled(false);
			pasteOnItem.setEnabled(false);
		}

		saveItem.setEnabled(changedSinceSaved);
		
		// disable all menus which can be disabled, having no enabled child menus or items
		disableIfUseless(editMenu);
	}
	
	/** 
	 * disable a menu item if it is useless, 
	 * i.e if it has no enabled menu items anywhere beneath it.
	 * 
	 * Recursively do the same for sub-menus
	 * 
	 * @param theMenu
	 */
	public static void disableIfUseless(JMenu theMenu)
	{
		String name = theMenu.getText(); // for debugging
		int uses = 0;
		
		for (Component child: theMenu.getMenuComponents()) 
			if (child instanceof JMenuItem)
		{
			String childName = ((JMenuItem)child).getText(); //for debugging
			// first try to disable the child
			if (child instanceof JMenu) disableIfUseless((JMenu)child);	
			// test if the child is still enabled
			if (((JMenuItem)child).isEnabled()) uses++;
		}
		
		theMenu.setEnabled(uses > 0);
	}
	
	/**
	 * 
	 * @param editableNode
	 */
	private void makeRemoveSlotMenu(FSNode editableNode)
	{
		removeSlotMenu.removeAll();
		
		for (Enumeration<String> en  = editableNode.getOwnSlotValues().keys();en.hasMoreElements();)
		{
			String slotName = en.nextElement();
			JMenuItem removeSlotItem = new JMenuItem(slotName);
			removeSlotMenu.add(removeSlotItem);
			removeSlotItem.addActionListener(new ActionListener() {
		        public void actionPerformed(ActionEvent ev) {
		        	removeSlotValue(slotName);
		        }
		    });
		}
	}

	
	// all editable attributes, on any node type
	static int NAMEATTRIBUTE = 0;
	static int DESCRIPTIONATTRIBUTE = 1;
	static int FREEORDER = 2;
	static int UNIFYORDER = 3;
	static int RESULT = 4;
	static String[] itemNames = {"Name","Description","Free Order Ranges","Unify Order","result"};
	
	// editable attributes by node type
	static int[] SCRIPTATTS = {NAMEATTRIBUTE,DESCRIPTIONATTRIBUTE,FREEORDER,UNIFYORDER}; // only 'script'
	static int[] SCENEATTS = {RESULT};
	static int[] OTHERATTS = {}; // for 'thing'
	static int[] NOATTS = {}; // for 'sound', 'prop', 'manner'
	
	/**
	 * make menu items to edit non-slot attributes on a node
	 */
	private void makeNodeValueMenuItems()
	{
		editOtherValueMenu.removeAll();
		
		if (selectedNode != null)
		{
			int[] atts;
			String type = selectedNode.getType();
			
			// attributes editable for script nodes
			if (type.equals("script")) atts = SCRIPTATTS;
			// attributes editable for scene nodes
			else if (type.equals("scene")) atts = SCENEATTS;
			// nodes with no editable attributes (usually with no child nodes)
			else if (type.equals("sound")|type.equals("prop")|type.equals("manner")) atts = NOATTS;
			//attributes editable for any other node type
			else atts = OTHERATTS;
			
			for (int pos = 0; pos < atts.length;pos++)
			{
				int editType = atts[pos];
				JMenuItem editItem = new JMenuItem(itemNames[editType]);
				editOtherValueMenu.add(editItem);
				editItem.addActionListener(new ActionListener() {
			        public void actionPerformed(ActionEvent ev) {
			        	editNodeAttribute(editType);
			        }
			    });
			}
		}
	}

	/**
	 * enable/disable Sound menu items, depending on what type of node is selected
	 * @param editableNode
	 */
	private void enableSoundItems(FSNode node)
	{
		soundBelowItem.setEnabled(false);
		soundAfterItem.setEnabled(false);
		soundBeforeItem.setEnabled(false);
		
		if (node != null)
		{
			soundBelowItem.setEnabled(node.getType().equals("script"));
			FSNode parent = node.getParent();
			if (parent != null)
			{
				boolean canAdd = (parent.getType().equals("script"));
				soundAfterItem.setEnabled(canAdd);
				soundBeforeItem.setEnabled(canAdd);
			}
		}
		
	}
	
	/**
	 * 
	 * @param node
	 */
	private void enableLinkItems(FSNode node)
	{
		addLinkStartItem.setEnabled(false);
		addLinkEndItem.setEnabled(false);
		removeLinkItem.setEnabled(false);
		
		if (node != null)
		{
			addLinkStartItem.setEnabled(true);
			
			boolean canEndLink = ((linkStartNode != null) &&  (node.getType().equals(linkStartNode.getType())));
			addLinkEndItem.setEnabled(canEndLink);
			
			boolean canRemoveLink = ((node.getSubsumingId() != null)|| (node.getSubsumedId() != null));			
			removeLinkItem.setEnabled(canRemoveLink);
		}
	}
	
	/**
	 * enable/disable Move menu items, depending on whether a node can move up or down
	 * @param editableNode
	 */
	private void enableMoveItems(FSNode node)
	{
		moveUpItem.setEnabled(false);
		moveDownItem.setEnabled(false);
		
		if (node != null)
		{
			FSNode parent = node.getParent();
			if (parent != null)
			{
				Vector<FSNode> siblings = parent.getChildren();
				if (siblings.size() > 1)
				{
					moveUpItem.setEnabled(true);
					moveDownItem.setEnabled(true);
					if (node.equals(siblings.get(0))) moveUpItem.setEnabled(false);
					if (node.equals(siblings.get(siblings.size() - 1))) moveDownItem.setEnabled(false);
				}
			}
		}
	}

	
    /**
     * make the sub-menus to add, edit or remove a learning example FS
     */
	private void makeExampleEditMenus()
    {
		// this method handles 5 sub-menus
		editLearningExampleMenu.removeAll();
		addLearningExampleMenu.removeAll();
		removeLearningExampleMenu.removeAll();
		addWordToLearnMenu.removeAll();
		editWordToLearnMenu.removeAll();
		removeWordToLearnMenu.removeAll();
		
		
    	for (LearningStage stage: model().getLearningStages())
    	{
       		String stageName = stage.getName();

       		// Edit > Learning Example> stage name
       		JMenu stageEditExampleMenu = new JMenu(stageName);
    		editLearningExampleMenu.add(stageEditExampleMenu);

    	    // add > Learning example > stage name
    	    JMenu stageAddExampleMenu = new JMenu(stageName);
			addLearningExampleMenu.add(stageAddExampleMenu);

			// Remove > Learning Example > stage name
    		JMenu stageRemoveExampleMenu = new JMenu(stageName);
    		removeLearningExampleMenu.add(stageRemoveExampleMenu);
    		
    		// Add > Word to Learn > stage name
    		JMenuItem stageAddWordLearnItem = new JMenuItem(stageName);
    	    addWordToLearnMenu.add(stageAddWordLearnItem);
    	    stageAddWordLearnItem.addActionListener(new ActionListener() {
		        public void actionPerformed(ActionEvent ev) {
		        	makeNewWordToLearn(stageName);
		        }
		    });
    	    
    		// Edit > Word to Learn > stage name
    		JMenu stageEditWordLearnMenu = new JMenu(stageName);
    		editWordToLearnMenu.add(stageEditWordLearnMenu);
    	    
    		// Remove > Word to Learn > stage name
    		JMenu stageRemoveWordLearnMenu = new JMenu(stageName);
    		removeWordToLearnMenu.add(stageRemoveWordLearnMenu);

     		
    		for (WordToLearn wl: stage.getLearnWords())
    		{
    			String wlName = wl.getWordName();
    			
    			// Edit > Word to Learn > Stage > word
				JMenuItem editWLItem  = new JMenuItem(wlName);
				stageEditWordLearnMenu.add(editWLItem);
				editWLItem.addActionListener(new ActionListener() {
			        public void actionPerformed(ActionEvent ev) {
			        	editWordToLearn(stageName,wlName);
			        }
			    });
    			
           		// Edit > Learning Example> stage name > word to learn
    			JMenu wlEditExMenu = new JMenu(wlName);
    			stageEditExampleMenu.add(wlEditExMenu);
    			
           		// Remove > Learning example > stage name > word to learn 
    			JMenu wlRemoveMenu = new JMenu(wlName);
    			stageRemoveExampleMenu.add(wlRemoveMenu);

    			// Add > Learning Example > stage name > word to learn				
				JMenuItem addExampleItem  = new JMenuItem(wlName);
				stageAddExampleMenu.add(addExampleItem);
				addExampleItem.addActionListener(new ActionListener() {
			        public void actionPerformed(ActionEvent ev) {
			        	makeNewExample(stageName,wlName);
			        }
			    });
				
    			// Remove > Word to learn > stage name > word to learn				
				JMenuItem removeWordLearnItem  = new JMenuItem(wlName);
				stageRemoveWordLearnMenu.add(removeWordLearnItem);
				removeWordLearnItem.addActionListener(new ActionListener() {
			        public void actionPerformed(ActionEvent ev) {
			        	vf.getAct().removeWordToLearn(stageName,wlName);
			        }
			    });
				
				   			
    			for (FSNode example: wl.getExamples())
    			{
    				String exName = example.getName();
    									
    				// Edit > Learning example > stage name > word to learn > example
    				JMenuItem exampleEditItem = new JMenuItem(exName);
    				wlEditExMenu.add(exampleEditItem);
    				exampleEditItem.addActionListener(new ActionListener() {
				        public void actionPerformed(ActionEvent ev) {
				        	setShownFSType(LEARNINGEXAMPLE);
				        	vf.getAct().showLearningExample(stageName,wlName,exName,EDIT);
				        }
				    });
    				
    				// Remove > Learning example > stage name > word to learn > example
    				JMenuItem exampleRemoveItem = new JMenuItem(exName);
    				wlRemoveMenu.add(exampleRemoveItem);
    				exampleRemoveItem.addActionListener(new ActionListener() {
				        public void actionPerformed(ActionEvent ev) {
				        	setShownFSType(LEARNINGEXAMPLE);
				        	vf.getAct().showLearningExample(stageName,wlName,exName,REMOVE);
				        }
				    });
    			}
    		}
    	}
    }

	

	
	/**
	 * 
	 * @param type
	 */
	private void makeNodeAddMenus(String type)
	{
		NodeType nt = model().getNodeType(type);
		addChildNodeMenu.removeAll(); // should not be necessary? Ineffective?
		for (String nodeType: nt.getChildNodeNames())
		{
			JMenuItem  additem = new JMenuItem(nodeType);
			addChildNodeMenu.add(additem);
			additem.addActionListener(new ActionListener() {
		        public void actionPerformed(ActionEvent ev) {
		        	doAddNode(nodeType);
		        }
		    });
		}
	}

	
	/**
	 * 
	 * @param type
	 */
	private void makeNodeChangeMenus(String type)
	{
		// make a slot sub-menu for each slot that can be on this node
		for (Enumeration<String> en = model().getSlots().keys();en.hasMoreElements();)
		{
			String slotName = en.nextElement();
			Slot slot = model().getSlots().get(slotName);
			if (slot.isAllowedNodeType(type))
			{
				JMenu slotMenu = new JMenu(slotName);
				editSlotMenu.add(slotMenu);
				for (SlotValue slotVal: slot.getValues())
					addSlotValueMenu(slotMenu,slot, slotVal);
			}
		}
	}
	
	/**
	 * recursively add menus and menu items for nested slot values
	 * @param slotMenu
	 * @param slot
	 * @param slotVal
	 */
	private void addSlotValueMenu(JMenu slotMenu,Slot slot, SlotValue slotVal)
	{
		// values with no sub-values; just add a menu item
		if (slotVal.getSubValues().size() == 0)
			addSlotValueItem(slotMenu,slot,slotVal);

		// if there are sub-values, add a menu, whose first item is the top value
		else
		{
			JMenu valueMenu = new JMenu(slotVal.getValue());
			slotMenu.add(valueMenu);
			addSlotValueItem(valueMenu,slot,slotVal);
			
			for (SlotValue subValue: slotVal.getSubValues())
				addSlotValueMenu(valueMenu, slot, subValue);
		}
	}

	/**
	 * 
	 * @param slotMenu
	 * @param slot
	 * @param slotVal
	 */
	private void addSlotValueItem(JMenu slotMenu,Slot slot, SlotValue slotVal)
	{
		final String value = slotVal.getValue();
		JMenuItem item = new JMenuItem(value);
		slotMenu.add(item);
		final String slotName = slot.getName();
		item.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	doEditSlot(slotName, value);
	        }
	    });
	}

	
//-------------------------------------------------------------------------------
//                    Edit Menu actions 
//-------------------------------------------------------------------------------

	
	//-------------------------------------------------------------------------------
	//	                   Editing a selected Feature Structure 
	//*-------------------------------------------------------------------------------

	/**
	 * all the specific changes to nodes, slots etc are action methods below.
	 * After any successful change, you need to do showChange().
	 */

/**
 * reflect the results of any editing change
 */
private void showChange()
{
	drawArea.repaint();
	changedSinceSaved = true;
	editing = true;
	refreshEditLanguageMenus(nodeToEdit,copiedNode,true);
}

	
	//-------------------------------------------------------------------------------
	//	                   Making a new Feature Structure (word or sentence)
	//*-------------------------------------------------------------------------------



/**
 * create a new FS to be edited
 */
private void doNewFS(int type) 
{
	editing = true;
	if (type == WORD)
	{
		// user chooses a name for the new FS
		String name = "";
		boolean newName = false;
		String title = "New name for a word feature structure";
		while (!newName)
		{
			name = Util.getShortUserInput(title,vf, name);
			newName = ((name.length() > 0) && (model().getWords().get(name) == null));
			if (!newName) title = "Word name '" + name + "' is already used.";
		}
		
		if (name.length() > 0) try
		{
			setEditState(WORD);
			
			FSNode newFS = new FSNode(model(),"script");
			newFS.setName(name);
			newFS.setEdited(true);
			model().addFS(newFS, WORD);
			
			drawArea.drawFS(false, newFS, null, null);
			changedSinceSaved = true; // creation of the new FS is a change
			
			// ready to edit the new script node
			nodeToEdit = newFS;
			refreshEditLanguageMenus(nodeToEdit,null,true);
		}
		catch (Exception ex) {ex.printStackTrace();}
	}
	
	else if (type == SENTENCE)
	{
		// get the proposed words in the sentence
		String sentenceName = "";
		boolean wordsOK = false;
		String title = "Words in the sentence, separated by spaces";
		while (!wordsOK)
		{
			sentenceName = Util.getShortUserInput(title,vf, sentenceName);
			if (sentenceName != null)
			{
				String nonWords = getNonWords(sentenceName);
				wordsOK = (nonWords.length() == 0);				
				if (!wordsOK) title = "Words " + nonWords + " are not yet in the language model.";
			}
			else wordsOK = true;
		}
		
		// the user has not cancelled and has entered some words, all valid
		if ((sentenceName != null) && (sentenceName.length() > 0)) try
		{
			setEditState(SENTENCE);
			
			// create a new sentence FS
			FSNode newFS = new FSNode(model(),"script");
			newFS.setName(sentenceName);
			newFS.setEdited(true);
			model().addFS(newFS, SENTENCE);
			
			// add the sounds of the words to the sentence
			addWordSounds(newFS,sentenceName,model());

			
			// draw the sentence
			drawArea.drawFS(false, newFS, null, null);
			changedSinceSaved = true; // creation of the new FS is a change
			refreshEditLanguageMenus(newFS, null, true);
			
			String question = "Create sentence meaning by understanding?";
			if (Util.getYesNo(question, vf))
			{
				understandMeaning(sentenceName,newFS);
			}
			
		}
		catch (Exception ex) {ex.printStackTrace();}
		
	}
}

/** 
 * check if a proposed sentence has any words not in the language model
 * @param words
 * @return
 */
private String getNonWords(String words)
{
	String nonWords = "";	
	StringTokenizer st = new StringTokenizer(words," ,.");	
	while (st.hasMoreTokens())
	{
		String word = st.nextToken();
		
		// if the word itself cannot be found
		if (model().getWords().get(word) == null) 
		{
			// check if the word has senses in brackets after the word
			Vector<FSNode> senses = model().getSenses(word);
			if (senses == null) nonWords = nonWords + word + " ";
		}
	}
	
	return nonWords;
}

/**
 *  add the sounds of the words to the sentence
 * @param newFS
 * @param sentenceName
 */
public static void addWordSounds(FSNode newFS,String sentenceName, LanguageModel model) throws Exception
{
	StringTokenizer st = new StringTokenizer(sentenceName," ,.");	
	while (st.hasMoreTokens())
	{
		String word = st.nextToken();
		FSNode wordFS = model.getWords().get(word);
		// iff you cannot  find the word itself, use the sounds of the first word sense (which are the same)
		if (wordFS == null) wordFS = model.getSenses(word).get(0);
		if (wordFS == null) throw new Exception("There is no word or word sense '" + word + "'");
		for (FSNode child: wordFS.getChildren()) if (child.getType().equals("sound"))
		{
			String sound = child.getSlotValue("sound");
			FSNode nextSound = new FSNode(model,"sound");
			nextSound.setSlotValue("sound", sound);
			newFS.addChild(nextSound);
		}
	}	
}

/**
 * 
 * @param sentenceName
 * @param newFS
 */
private void understandMeaning(String sentenceName, FSNode newFS) throws Exception
{
	// initial unify order is computed automatically - user may wish to recompute
	String unifyOrder = vf.getAct().makeUnifyOrderString(sentenceName);

	boolean validOrder = false;
	String title = "Define unification order for understanding";	
	while(!validOrder)
	{
		unifyOrder = Util.getShortUserInput(title,vf, unifyOrder);
		if (unifyOrder != null)
		{
			String err = orderError(sentenceName,unifyOrder);
			if (err.length() > 0) title = "Error in unify order: " + err;
			else validOrder = true;
		}
		else validOrder = true; // but cancelled		
	}
	
	if ((unifyOrder != null) && (unifyOrder.length() > 0)) try
	{
		newFS.setUnifyOrder(unifyOrder);
		
		Vector<Vector<FSNode>> understandResult = vf.getAct().generateOrUnderstandAllSteps(newFS, false);
		
		// all steps of unification succeeded
		if (understandResult.size() == newFS.getUnifyWords().size())
		{
			// final result is the 3rd FS on the last row
			FSNode finalResult = understandResult.get(understandResult.size() - 1).get(2);

			// move the result meaning structure into the sentence FS being created
			Vector<FSNode> children = finalResult.getChildren();
			FSNode meaningScene = children.get(children.size() -1);
			newFS.addChild(meaningScene);

			// redraw the sentence, with meaning
			drawArea.drawFS(false, newFS, null, null);
			changedSinceSaved = true; // creation of the new FS is a change
			refreshEditLanguageMenus(newFS, null, false);
		}
		
		// some step failed
		else
		{
			title = ("Unification failed after the first " + understandResult.size() + " steps");
			boolean keep = Util.getYesNo(title, vf);
		}
	}
	catch (Exception ex) {ex.printStackTrace();}
}

/** 
 * check that the unify order is a permutation of the words in the sentence
 * @param sentenceName
 * @param unifyOrder
 * @return
 */
private String orderError(String sentenceName,String  unifyOrder)
{
	String err = "";
	
	/* make a table of each word with its number of occurrences. 
	 * sentenceName has words without senses in brackets */
	Hashtable<String,Integer> words = new Hashtable<String,Integer>();
	StringTokenizer st = new StringTokenizer(sentenceName," ,.");	
	while (st.hasMoreTokens())
	{
		String word = st.nextToken();
		Integer iv = words.get(word);
		if (iv == null) iv = Integer.valueOf(0);
		iv = Integer.valueOf(iv.intValue() + 1);
		words.put(word, iv);		
	}
	
	/* check each word in the unify order is in the table; if so, decrement the count for the word.
	 * unify order has words with senses in brackets*/
	StringTokenizer ut = new StringTokenizer(unifyOrder," ,.");	
	while (ut.hasMoreTokens())
	{
		String word = ut.nextToken();
		Integer iv = words.get(word);
		if (iv == null) 
		{
			// a word sense may need to be reduced to a bare word
			StringTokenizer su = new StringTokenizer(word,"()");
			word = su.nextToken();
			iv = words.get(word);
			if (iv == null) err = err + word + " unknown; ";
		}
		
		if (iv != null)
		{
			// decrement the count for a known word
			iv = Integer.valueOf(iv.intValue() - 1);
			words.put(word, iv);		
		}
	}
	
	// now all word counts (for bare words, without senses) should be zero
	for (Enumeration<String> en = words.keys(); en.hasMoreElements();)
	{
		String word = en.nextElement();
		Integer iv = words.get(word);
		if (iv.intValue() != 0) err = err + word + " missing; ";
	}
	
	return err;
}

//-------------------------------------------------------------------------------
//	                   Links
//*-------------------------------------------------------------------------------

/**
 * set up a node as a link start, and choose a new link id (unique for the whole FS)
 * makes no change to the feature structure, yet.
 */
private void doAddLinkStart()
{
	linkStartNode = selectedNode;
	linkId = selectedNode.getNextLinkId();
	
	refreshEditLanguageMenus(selectedNode, copiedNode, true);
}


/**
 * add a link to a feature structure.
 */
private void doAddLinkEnd()
{
	linkStartNode.setSubsumingId(linkId);
	selectedNode.setSubsumedId(linkId);
	
	linkStartNode = null;
	linkId = null;
	
	showChange();
}

/**
 * find the link id of any link onto the selected node, 
 * and remove that link from the whole FS
 */
private void doRemoveLink()
{
	String id = selectedNode.getSubsumingId();
	if (id == null) id = selectedNode.getSubsumedId();
	if (id != null)
	{
		// rem should be 2, for removing both ends
		int rem = selectedNode.getRootAncestor().removeLink(id);
		showChange();
	}
	
}

	//-------------------------------------------------------------------------------
	//	                   Learning Examples
	//*-------------------------------------------------------------------------------


/**
 * 
 * @param stageName
 */
private void makeNewWordToLearn(String stageName)
{
	editing = true;
	Util.message("making word to learn in stage " + stageName);
			
	try
	{
		LearningStage stage = model().getLearningStage(stageName);
		boolean newName = false;
		String wordName = "";
		String title = "Name for new word to learn";
		while (!newName)
		{
			wordName = Util.getShortUserInput(title, vf, wordName);
			if (stage.getWordLearn(wordName) != null)
				title = "Word " + wordName + " is already in learning stage " + stageName;
			else newName = true;
		}
		
		// make a new word to learn, with an empty feature structure,  and add it to the learning stage
		WordToLearn wl = new WordToLearn(stage, wordName);	
		stage.addWord(wl);
		
		// feedback to the user is in the text area
		vf.getAct().explain("Added new word to learn '" + wl.getWordName() + "' in learning stage '" + stage.getName() + "'");
		
		changedSinceSaved = true;
		refreshEditLanguageMenus(null,null,false);
	}
	catch (Exception ex) {ex.printStackTrace();}
}

/**
 *  
 * @param stageName
 * @param wlName
 */
void editWordToLearn(String stageName,String wlName)
{
	editing = true;
	LearningStage stage = model().getLearningStage(stageName);
	WordToLearn wl = stage.getWordLearn(wlName);
	FSNode toEdit = wl.getWordFS();
	
	if (toEdit != null)
	{
		vf.getAct().showTheFS(toEdit, WORDTOLEARN, true);
	}
	else Util.message("Word to learn " + wlName + " has no target feature structure.");
 
}


/**
 * 
 * @param stageName
 * @param wordName
 */
private void makeNewExample(String stageName,String wordName)
{
	editing = true;
	Util.message("Making new example to learn word " + wordName + " in stage " + stageName);
	
	try
	{
		LearningStage stage = model().getLearningStage(stageName);
		WordToLearn wl = stage.getWordLearn(wordName);
		
		boolean newName = false;
		String exampleName = "";
		String title = "Name for new learning example to learn word " + wordName;
		while (!newName)
		{
			exampleName = Util.getShortUserInput(title, vf, exampleName);
			if (Util.inVector(exampleName, wl.getExampleNames()))
				title = "Learning example " + exampleName + " is already used to learn word " + wordName;
			else newName = true;
		}
		
		FSNode lex = new FSNode(model(),"script");
		lex.setName("Ex: " + exampleName);
		wl.addExample(lex);
		
		showExampleForEditing(lex, exampleName);

	}
	catch (Exception ex) {ex.printStackTrace();}

}

/**
 * show a learning example for editing
 * @param lex
 */
private void showExampleForEditing(FSNode lex, String exampleName)
{
	editing = true;
	lex.setEdited(true);
	drawArea.drawFS(false,lex,null, null);
	setEditState(LEARNINGEXAMPLE);
	refreshEditLanguageMenus(lex,null,true);
}

/**
 * 
 * @param stageName
 * @param wordName
 * @param exampleName
 */
private void editExample(String stageName,String wordName,String exampleName)
{
	Util.message("editing example " + exampleName + " used to learn word " + wordName + " in stage " + stageName);
	
	try
	{
		LearningStage stage = model().getLearningStage(stageName);
		WordToLearn wl = stage.getWordLearn(wordName);			
		FSNode lex = wl.getExample(exampleName);
		showExampleForEditing(lex,exampleName);
	}
	catch (Exception ex) {ex.printStackTrace();}

}


	//-------------------------------------------------------------------------------
	//	                    Copy & paste
	//*-------------------------------------------------------------------------------


private void doCopy() 
{
	// make a clone of  the selected node
	copiedNode = new FSNode(selectedNode);
	
	// paste items may be enabled or not ,depending on the type of the nodes
	refreshEditLanguageMenus(nodeToEdit,copiedNode,true);
}


/**
 * Paste a node below the selected node
 */
private void doPasteBelow() 
{
	//find out which FS are currently drawn
	Vector<FSNode> drawnNow = drawArea.getLastRow();
	FSNode edited = drawnNow.get(0);
	FSNode copiedFrom = null;
	if (drawnNow.size() > 1) copiedFrom = drawnNow.get(1);
	
	// checks
	if (nodeToEdit == null) Util.message("no node to edit");
	else if (copiedNode == null) Util.message("no copied node");
	else Util.message("sel: " + selectedNode.getType() + "; cop " + copiedNode.getType() + "\n");
	
	// do the paste on the selected node - make a clone of the copied node, to change FS 'edited'
	nodeToEdit.addChild(new FSNode(copiedNode));
	
	// sort out the names of all new descendant nodes
	nodeToEdit.setName(nodeToEdit.getName());
	
	// redraw the altered FS and the FS the node was copied from
	drawArea.drawFS(false, edited, copiedFrom, null);
	changedSinceSaved = true;
	
	// write out the result
	edited.write("");
	
	// the copied node is no longer available for pasting
	copiedNode = null;
}

/**
 * 
 */
private void doPasteOn() 
{
	if ((selectedNode != null) && (copiedNode != null))
	{
		// paste a clone
		FSNode toPaste = new FSNode(copiedNode);
		boolean scriptPaste = (selectedNode.getType().equals("script"));

		// top level script paste - replace everything except the name
		if (scriptPaste) 
		{
			String name = selectedNode.getName();
			toPaste.setName(name);
			savePastedScript(selectedNode,toPaste);
			
			//change the drawn FS, keeping the FS that was copied from
			drawArea.drawFS(false, toPaste, copiedNode, null);
		}
		// other nodes - replace the previous node as a child of its parent node, at the same position
		else
		{
			FSNode parent = selectedNode.getParent();
			Vector<FSNode> siblings = parent.getChildren();
			int index = siblings.indexOf(selectedNode);
			if (index > -1)
			{
				siblings.remove(index);
				siblings.insertElementAt(toPaste, index);				
			}
			else Util.message("Cannot remove node being pasted on");
		}
		
		showChange();
	}
}

/**
 * if a top level Script node has been pasted, 
 * it needs to be saved at the appropriate place in the language model
 * @param oldScript
 * @param newScript
 */
private void savePastedScript(FSNode oldScript, FSNode newScript)
{
	WordToLearn wlOwner = oldScript.getWordLearnToLearnParent();
	WordToLearn wlParent = oldScript.getWordLearnExampleParent();

	// feature structure to be learnt, for a word to learn
	if (wlOwner != null)
	{
		wlOwner.setWordFS(newScript);
	}
	
	
	// Learning example -save at the same position as the example it replaces
	else if (wlParent != null)
	{
		Vector<FSNode> siblings = wlParent.getExamples();
		int index = siblings.indexOf(oldScript);
		if (index > -1)
		{
			siblings.remove(index);
			siblings.insertElementAt(newScript, index);				
		}
		else Util.message("Cannot remove learning example being pasted on");
	}

	// Word
	else if (getShownFSType() == WORD)
	{
		// replace the old Feature Structure with the same name
		model().getWords().put(newScript.getName(), newScript);
	}

	// Sentence
	else if (getShownFSType() == SENTENCE)
	{
		// replace the old Feature Structure with the same name
		model().getSentences().put(newScript.getName(), newScript);
	}
	
	else Util.message("Editor does not know what to do with the Paste On result");
	
	showChange();
}

	
	//-------------------------------------------------------------------------------
	//	                    Add a child node
	//*-------------------------------------------------------------------------------


/**
 * 
 * @param nodeType
 */
private void doAddNode(String nodeType)
{
	editing = true;
	try
	{
		FSNode node = new FSNode(model(), nodeType);
		nodeToEdit.addChild(node);
		
		showChange();
		
	}
	catch (Exception ex) {}
}
	
	//-------------------------------------------------------------------------------
	//	                    Edit slot values, or remove slots
	//*-------------------------------------------------------------------------------


/**
 * 
 * @param slotName
 * @param value
 */
private void doEditSlot(String slotName, String value)
{
	editing = true;
	nodeToEdit.setSlotValue(slotName, value);
	showChange();
}
	
	//-------------------------------------------------------------------------------
	//	                   Change values of other node attributes 
	//*-------------------------------------------------------------------------------

	/**
	 * edit various types of attribute on the selected node
	 * @param editType
	 */
	private void editNodeAttribute(int editType)
	{
		if (selectedNode != null)
		{
			boolean changed = false;
			if (editType == NAMEATTRIBUTE)  // only for top level script nodes
			{
				String oldName = selectedNode.getName();
				String newName = Util.getShortUserInput("Name for Feature Structure", vf, oldName);
				if (newName != null)
				{
					selectedNode.setName(newName);
					changed = true;
					// FIXME - changing the name of a sentence or a word changes its storage; and the new name must not be taken
				}
			}
			
			if (editType == DESCRIPTIONATTRIBUTE) // only for top level script nodes
			{
				String oldDescription = selectedNode.getDescription();
				if (oldDescription == null) oldDescription = "";
				vf.getAct().explain(oldDescription);
				
				changed = true; //enable save, even if the user subsequently does nothing.
				vf.startEdit(VFrame.FSDESCRIPTION,selectedNode);		
			}
			
			else if (editType == FREEORDER)
			{
				boolean orderOK = false;
				String title = "Free Order Ranges";
				String previous = selectedNode.freeOrderString();
				
				// keep going until the user enters valid order constraints, or gives up
				while (!orderOK)
				{
					String newOrder = Util.getShortUserInput(title, vf,previous) ;
					if (newOrder != null)
					{
						try
						{
							Vector<int[]> order = selectedNode.getFreeOrderGroups(newOrder);
							selectedNode.setFreeOrder(order);
							changed = true;
							orderOK = true;
						}
						// tell the user what is wrong, and let him try again
						catch (Exception ex) {title = ex.getMessage();previous = newOrder;}
					}
					//user gives up
					else if (newOrder == null) orderOK = true;					
				}				
			}
			
			else if (editType == UNIFYORDER)
			{
				String title = "Unify Order of words";
				String previous = selectedNode.getUnifyOrder();
				String missing = "something";
				String newUnifyOrder = null;
				
				// keep asking until no words are missing from the dictionary
				while (missing.length() > 0)
				{
					newUnifyOrder = Util.getShortUserInput(title, vf, previous);
					if (newUnifyOrder != null)
					{
						missing = missingWords(newUnifyOrder);
						title = "Words that cannot be found: " + missing;
						previous = newUnifyOrder;
					}
					else if (newUnifyOrder == null) missing = "";
				}
					
				// user did not cancel
				if (newUnifyOrder != null)
				{
					selectedNode.setUnifyOrder(newUnifyOrder);
					changed = true;
				}
				
			}
			
			else if (editType == RESULT)
			{
				String oldResult = Boolean.toString(selectedNode.isResult());
				String newResult = Util.getShortUserInput("Result flag", vf, oldResult);
				if (newResult != null)
				{
					if (newResult.equalsIgnoreCase("true")) selectedNode.setIsResult(true);
					else selectedNode.setIsResult(false);
					changed = true;
				}
				
			}
			
			if (changed) showChange();
		}
	}
	
	/**
	 * to check that a unify order contains only words in the dictionary - 
	 *  words with senses will count , but word stems do not
	 * @param unifyOrder
	 * @return
	 */
	String missingWords(String unifyOrder)
	{
		String missing = "";
		StringTokenizer st = new StringTokenizer(unifyOrder," ");
		while (st.hasMoreTokens())
		{
			String word = st.nextToken();
			if (model().getWords().get(word) == null) missing = missing + word + " ";
		}
		
		return missing;
	}
	
	//-------------------------------------------------------------------------------
	//	                   Mode Up/Down 
	//*-------------------------------------------------------------------------------


private void doMove(boolean moveUp)
{
	if (selectedNode != null)
	{
		FSNode parent = selectedNode.getParent();
		if (parent != null)
		{
			Vector<FSNode> siblings = parent.getChildren();
			int nc = siblings.size();
			if (nc > 1) //can only make a swap if there is more than one sibling
			{
				Vector<FSNode> newChildren = new Vector<FSNode>();
				for (int i = 0;  i < nc; i++)
				{
					boolean swapped = false;
					FSNode child = siblings.get(i);
					
					// only look for swaps before the last child
					if (i < nc - 1)
					{
						FSNode nextChild = siblings.get(i + 1);
						// conditions for a swap
						boolean doSwap = ((moveUp) && (selectedNode.equals(nextChild))) || ((!moveUp) && (selectedNode.equals(child)));
						if (doSwap)
						{
							// make the swap, and ensure the next child does not repeat
							newChildren.add(nextChild);
							newChildren.add(child);
							swapped = true;
							i++;
						}
					}
					
					// if the previous two nodes were swapped, do not add one of them again; or add it next time round the loop
					if (!swapped)  newChildren.add(child);					
				}
				
				parent.setChildren(newChildren);
				showChange();
				
			} // end of if (nc > 1)
		} // end of if parent != null
	} // end of it(selectedNode != null
	
}
	
	//-------------------------------------------------------------------------------
	//	                   Sounds 
	//*-------------------------------------------------------------------------------



/**
 * Enter a sequence of sounds
 * @param type - values
 * 0 - add sounds below script node
 * 1 - add sounds before sound node
 * 2 - add sounds after sound node
 */
private void doSound(int type)
{
	boolean validSounds = false;
	boolean cancelled = false;
	String sounds = "";
	String title = "Enter sounds, separated by spaces";
	
	// keep going until all sounds are valid in the model, or the user cancels
	while ((!validSounds) && (!cancelled))
	{
		 sounds = Util.getShortUserInput(title, vf,sounds);
		 if (sounds == null) cancelled = true;
		 else if (sounds.length() == 0) cancelled = true;
		 else
		 {
				Vector<String> banned = bannedSounds(sounds);
				if (banned.size() > 0)
				{
					title = "Sounds ";
					for (String b: banned) title = title + "'" + b + "' ";
					title = title + " are not in the language model.";
				}
				else validSounds = true;
		 }
	}
	
	if (cancelled) return;
	
	// convert the sounds to a list of FeatureStructures
	Vector<FSNode> allSoundFS = new Vector<FSNode>();	
	try
	{
		StringTokenizer st = new StringTokenizer(sounds," ");
		while (st.hasMoreTokens())
		{
			String sound = st.nextToken();
			FSNode soundFS = new FSNode(model(),"sound");
			soundFS.setSlotValue("sound", sound);
			allSoundFS.add(soundFS);
		}
		
	}
	catch (Exception ex) {return;};
	
	// add below a script - after all existing child nodes
	if (type == 0)
	{
		for (FSNode soundFS: allSoundFS) nodeToEdit.addChild(soundFS);
	}
	
	// add below a script, before (type = 2) or after (type = 3) a selected child node
	else if (type > 0)
	{
		FSNode parent = nodeToEdit.getParent();
		Vector<FSNode> oldChildren = parent.getChildren();
		Vector<FSNode> newChildren = new Vector<FSNode>();
		
		for (FSNode child: oldChildren)
		{
			if (child.equals(nodeToEdit))
			{
				// add new FS before
				if (type == 1)
				{
					for (FSNode added: allSoundFS) newChildren.add(added);
					newChildren.add(child);
				}
				
				// add new FS after
				if (type == 2)
				{
					newChildren.add(child);
					for (FSNode added: allSoundFS) newChildren.add(added);
				}
			}
			
			else newChildren.add(child);
		}
		
		parent.setChildren(newChildren);
	}
			
	showChange();
}

/**
 * detect sounds which are not in the model
 * @param sounds
 * @return
 */
private Vector<String> bannedSounds(String sounds)
{
	Vector<String> banned = new Vector<String>();
	StringTokenizer st = new StringTokenizer(sounds," ");
	while (st.hasMoreTokens())
	{
		String sound = st.nextToken();
		boolean valid = model().getSlot("sound").isAllowedValue(sound);
		if (!valid) banned.add(sound);
	}
	
	return banned;
}
	
	//-------------------------------------------------------------------------------
	//	                   Remove a Node or a Feature Structure  or a slot value
	//*-------------------------------------------------------------------------------

/**
 * remove a node from the feature structure
 */
private void doRemoveNode()
{
	if (nodeToEdit != null)
	{
		// cannot remove the top 'Script' node
		if (!nodeToEdit.getType().equals("script"))
		{
			// remove both ends of all links which have one end on or below the removed node, before removing it
			Hashtable<String,String> linkIds = new Hashtable<String,String>();
			nodeToEdit.addLinkIds(linkIds);
			for (Enumeration<String> en = linkIds.keys(); en.hasMoreElements();)
			{
				String linkId = en.nextElement();
				nodeToEdit.getRootAncestor().removeLink(linkId);
			}
			
			FSNode parent = nodeToEdit.getParent();

			Vector<FSNode> newChildren  = new Vector<FSNode>();			
			for (FSNode child: parent.getChildren())
				if (!nodeToEdit.equals(child)) newChildren.add(child);
			
			parent.setChildren(new Vector<FSNode>());
			for (FSNode child:newChildren) parent.addChild(child);
			
			showChange();
		}
	}
}

/**
 * 
 * @param slotName
 */
private void removeSlotValue(String slotName)
{
	if (nodeToEdit != null)
	{
		nodeToEdit.removeSlotValue(slotName);
		showChange();
	}
}

/**
 * remove a whole Feature Structure
private void doRemoveFS()
{
	if (nodeToEdit != null)
	{
	
	}
}
 */
	
	//-------------------------------------------------------------------------------
	//	                   Save
	//*-------------------------------------------------------------------------------


/**
 * save all FS of any type
 */
private void doSave() 
{
	try
	{
		model().saveLanguageFiles();
		changedSinceSaved = false;
		FSNode beingEdited = drawArea.getDrawnFS(0);
		refreshEditLanguageMenus(beingEdited, null, false);
	}
	catch (Exception ex) {ex.printStackTrace();Util.message(ex.getMessage());}
}

	
	//-------------------------------------------------------------------------------
	//	                   Finish Editing
	//*-------------------------------------------------------------------------------


private void doFinish() 
{
	setEditState(NONE);
	editing = false;
	
	if ((changedSinceSaved) && (Util.getYesNo("Save changes?", vf))) doSave();

	drawArea.clear();
		
	finishItem.setEnabled(false);
}


//-------------------------------------------------------------------------------
//	                   
//-------------------------------------------------------------------------------







}
