package displayPack;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.swing.JMenu;
import javax.swing.JMenuItem;

import modelPack.FSNode;
import modelPack.LanguageModel;
import modelPack.LearningStage;
import modelPack.NodeType;
import modelPack.Slot;
import modelPack.SlotValue;
import modelPack.WordToLearn;
import utilPack.Util;

public class EditModel {
	
	private DrawArea drawArea;

	private VFrame vf;
	
	private Actions act;
	
	private LanguageModel model() {return vf.getModel();} 
	
	// variables of different types to be edited
	
	private Slot selectedSlot = null;
	
	private SlotValue selectedSlotValue  = null;
	
	private NodeType selectedNodeType = null;
	public void setSelectedNodeType(NodeType nt) {selectedNodeType = nt;}
	
	private Topic selectedTopic = null;
	public void setSelectedTopic(Topic top) {selectedTopic = top;}
	
	public static int NONE = -1;
	public static int MODEL = 1;
	
	private int editState = NONE;
	
	/**
	 * set the new edit state
	 * @param state
	 */
	public void setEditState(int state) {editState = state;}

	// true if the shown FS can be edited
	public boolean canEdit() {return (editState == MODEL);}

	// true if the shown FS has been changed since the last save
	private boolean modelChangedSinceSaved = false;
	public void setChangedSinceSaved(boolean changed) {modelChangedSinceSaved = changed;}

	
	JMenu editModelMenu;
	JMenu viewMenu = new JMenu("View");
		JMenu viewSlotMenu = new JMenu("Slots");
		JMenu viewNodeTypeMenu = new JMenu("Node Types");
	JMenu editMenu = new JMenu("Edit");
		JMenuItem editSlotValueItem = new JMenuItem("Slot value");
		JMenu editSlotMenu = new JMenu("Slot");
		JMenu editNodeTypeMenu = new JMenu("Node Type");
		JMenuItem editTopicItem = new JMenuItem("Topic");
	JMenu addMenu = new JMenu("Add");
		JMenuItem addValueItem = new JMenuItem("Slot value");
		JMenuItem addSlotItem = new JMenuItem("Slot");
		JMenuItem addNodeTypeItem = new JMenuItem("Node Type");
	JMenu removeMenu = new JMenu("Remove");
		JMenuItem removeValueItem = new JMenuItem("Slot value");
		JMenuItem removeSlotItem = new JMenuItem("Slot");
		JMenuItem removeNodeTypeItem = new JMenuItem("Node Type");
	JMenuItem checkLanguageItem = new JMenuItem("Check Language");
	JMenuItem saveItem = new JMenuItem("Save");
	JMenuItem finishItem = new JMenuItem("Finish");


	  //------------------------------------------------------------------------------------------------
	  //                              constructor and startup
	  //------------------------------------------------------------------------------------------------
	
	
	public EditModel(VFrame vf, DrawArea drawArea)
	{
		this.vf = vf;
		act = vf.getAct();
		this.drawArea = drawArea;
	}
	

	  //------------------------------------------------------------------------------------------------
	  //                              setting up the edit menus
	  //------------------------------------------------------------------------------------------------
	
	/**
	 * 
	 */
	public void setUpEditMenu(JMenu editModelMenu)
	{
		this.editModelMenu = editModelMenu;
		editModelMenu.add(viewMenu);
			viewMenu.add(viewSlotMenu);
			viewMenu.add(viewNodeTypeMenu);
		editModelMenu.add(editMenu);
			editMenu.add(editSlotValueItem);
			editMenu.add(editSlotMenu);
			editMenu.add(editNodeTypeMenu);		
			editMenu.add(editTopicItem);		
		editModelMenu.add(addMenu);
			addMenu.add(addValueItem);
			addMenu.add(addSlotItem);
			addMenu.add(addNodeTypeItem);
	    editModelMenu.add(removeMenu);
	    	removeMenu.add(removeValueItem);
	    	removeMenu.add(removeSlotItem);
	    	removeMenu.add(removeNodeTypeItem);
	    editModelMenu.add(checkLanguageItem);
	    editModelMenu.add(saveItem);
	    editModelMenu.add(finishItem);
	    
	    editSlotValueItem.setEnabled(false);
	    editTopicItem.setEnabled(false);
	    addValueItem.setEnabled(false);
	    removeSlotItem.setEnabled(false);
	    removeValueItem.setEnabled(false);
	    removeNodeTypeItem.setEnabled(false);
	    saveItem.setEnabled(false);
	    finishItem.setEnabled(false);

	    addValueItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	addValueTo();
	        }
	    });
    	
    	
	    addSlotItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	addSlot();
	        }
	    });
    	
    	
	    addNodeTypeItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	addNodeType();
	        }
	    });
     	

	    removeValueItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	removeValue();
	        }
	    });
    	
    	
	    removeSlotItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	removeSlot();
	        }
	    });
    	
    	
	    removeNodeTypeItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	removeNodeType();
	        }
	    });
     	
    	
	    editSlotValueItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	editSlotValue();
	        }
	    });
    	
	    editTopicItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	editTopic();
	        }
	    });
    	
    	saveItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	doSave();
	        }
	    });
    	
    	checkLanguageItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	checkLanguage();
	        }
	    });
    	
    	finishItem.addActionListener(new ActionListener() {
	        public void actionPerformed(ActionEvent ev) {
	        	doFinish();
	        }
	    });		
	}
	

/**
 * called when the user selects any node on any feature structure
 * @param fs
 */
public void notifySelect(SlotValue sv)	
{
	selectedSlotValue = sv;
	selectedNodeType = null;
	
	if (sv != null) selectedSlot  = sv.getSlot();
	 

	// refresh all menus for the selected node; false = not atStart
	refreshEditModelMenus(false,null);
}

	
	/**
	 * enable or disable edit menu items, and create those that depend on data 
	 */
	public void refreshEditModelMenus(boolean atStart,MenuElement modelElement)
	{
		//usually there is no need to send the refreshed edit menu to the web client
		if (modelElement == null) modelElement = new MenuElement("dummy");
		
		
		if (!atStart)
		{
			viewNodeTypeMenu.removeAll();
			viewSlotMenu.removeAll();
			editNodeTypeMenu.removeAll();
			editSlotMenu.removeAll();
		}
		
		// add menu items to view node types and slots - always enabled
		MenuElement nodeElement = new MenuElement("Node Types");
		modelElement.addChild(nodeElement);		
		addNodeTypeItems(model().getNodeTypes(), viewNodeTypeMenu,act,false,nodeElement);
		
		// MenuElement slotElement = new MenuElement("Slots");
		// modelElement.addChild(slotElement);
		/* for the web client, to make all menus two-level, the slot alphabetical letter groups are added direct to the model menu,
		 * not through an extra 'slots'  submenu */
		addSlotItems(model().getSlots(), viewSlotMenu,act,false,modelElement);
		
		// add menu items to edit node types and slots - always enabled
		addNodeTypeItems(model().getNodeTypes(), editNodeTypeMenu,act,true,null);
		addSlotItems(model().getSlots(), editSlotMenu,act,true,null);
		
		editSlotValueItem.setEnabled(selectedSlotValue != null);
		editTopicItem.setEnabled(selectedTopic != null);
		addValueItem.setEnabled(selectedSlotValue != null);

		removeValueItem.setEnabled(selectedSlotValue != null);
		removeSlotItem.setEnabled(selectedSlot != null);
		removeNodeTypeItem.setEnabled(selectedNodeType != null);
		
		saveItem.setEnabled(modelChangedSinceSaved);
		finishItem.setEnabled(canEdit());
		
	}
	
	/**
	 * 
	 * @param nodeTypes
	 * @param nodeMenu
	 * @param act
	 * @param doEdit; whether true or false, show the node type in the text area
	 */
	private void addNodeTypeItems(Hashtable<String,NodeType> nodeTypes, JMenu nodeMenu,Actions act,boolean doEdit,MenuElement nodeElement)
	{
		if (nodeElement == null) nodeElement = new MenuElement("dummy");
		
		for (Enumeration<String> en = nodeTypes.keys();en.hasMoreElements();)
		{
			String nodeName = en.nextElement();
			JMenuItem nodeItem = new JMenuItem(nodeName);
			
			// to send to the client
			MenuElement nEl = new MenuElement(nodeName);
			nodeElement.addChild(nEl); 
			
			final EditModel em = this;
			nodeItem.addActionListener(new ActionListener() {
		    	final NodeType nType = nodeTypes.get(nodeName);
		        public void actionPerformed(ActionEvent ev) {
		        	selectedNodeType = nType;
		        	act.showNode(nType,doEdit,em);
		        	if (doEdit) editNodeType();
		        }
		    });
			nodeMenu.add(nodeItem);
		}
	}


	/**
	 * 
	 * @param slots
	 * @param addTo
	 * @param act
	 * @param doEdit
	 */
	private void addSlotItems(Hashtable<String,Slot> slots, JMenu addTo, Actions act, boolean doEdit, MenuElement slotElement)
	{
		if (slotElement == null) slotElement = new MenuElement("dummy");
		
		// Hashtable of all slots, keyed by initial capital letter
		Hashtable<String,Vector<String>> names = new Hashtable<String,Vector<String>>();
		
		for (Enumeration<Slot> en = slots.elements(); en.hasMoreElements();)
		{
			Slot slot = en.nextElement();
			String slotName = slot.getName();
			// upper case initial letter
			String initial = slotName.substring(0,1).toUpperCase();
			
			Vector<String> nl = names.get(initial);
			if (nl == null) nl = new Vector<String>();
			nl.add(slotName);
			names.put(initial,nl);
		}
		
		for (int i = 0; i < VFrame.alphabet.length; i++)
		{
			String initial = VFrame.alphabet[i];
			Vector<String> nl = names.get(initial);
			if (nl != null)
			{
				JMenu initMenu = new JMenu(initial);
				addTo.add(initMenu);
				
				// to send to the client
				MenuElement initEl = new MenuElement(initial);
				slotElement.addChild(initEl); 

				for (String slotName: nl)
				{
					JMenuItem slotItem = new JMenuItem(slotName);
					
					// to send to the client
					MenuElement sEl = new MenuElement(slotName);
					initEl.addChild(sEl); 

					// draw a slot in the drawing area if it is to be edited
					if (doEdit)
					{
						slotItem.addActionListener(new ActionListener() {
					    	final Slot s = slots.get(slotName);
					        public void actionPerformed(ActionEvent ev) {
					        	drawSlot(s,true);
					        }
					    });
					}
					
					// if a slot is only to be viewed, show it in the text area
					else if (!doEdit)
					{
						slotItem.addActionListener(new ActionListener() {
					    	final Slot s = slots.get(slotName);
					        public void actionPerformed(ActionEvent ev) {
					        	act.showSlot(s);
					        }
					    });
					}
					
					// add the slot menu item under the correct initial
					initMenu.add(slotItem);
				}
			}
		} // end of loop over alphabet
		
		
		// to make the alpha sub-menu look wider
		String widener = "---------";
		JMenu wideMenu = new JMenu(widener);
		addTo.add(wideMenu);

	}

	
	
//-------------------------------------------------------------------------------
//	                       Edit actions for Node types 
//-------------------------------------------------------------------------------


/**
 * 
 */
private void addNodeType()
{
	setEditState(MODEL);
	
	String name = Util.getShortUserInput("Name of new node type", vf, "");
		
	// do nothing if the user has cancelled
	if (name != null)
	{
		// do nothing if the name is not new
		if (model().getNodeType(name) == null)
		{
			selectedNodeType = new NodeType(model(),name);
			model().addNodeType(selectedNodeType);
			modelChangedSinceSaved = true;
			refreshEditModelMenus(false,null);
			
			editNodeType();
		}
		else  act.explain("Node name " + name + " is aready used.");				
	}	
}

/**
 * remove the selected node type
 */
private void removeNodeType()
{
	if (selectedNodeType !=  null)
	{
		model().getNodeTypes().remove(selectedNodeType.getName());
		
		modelChangedSinceSaved = true;
		refreshEditModelMenus(false,null);
	}
}

/**
 * 
 */
private void editNodeType()
{
	Vector<String> prev = selectedNodeType.getChildNodeNames();
	String previous = "";
	for (String p: prev) previous = previous + p + " ";
	
	String childNodeTypes = getAllowedNodeTypes("Child node types",previous); 
	if (childNodeTypes == null) childNodeTypes = previous;
	selectedNodeType.setChildNodeTypes(childNodeTypes);	
	act.showNode(selectedNodeType, false, this);
	
	modelChangedSinceSaved = true;
	refreshEditModelMenus(false,null);
	
	// editing the description is always done last, because control is not returned
	if (Util.getYesNo("Edit description of node type?",vf))
	{
		act.explain(selectedNodeType.getDescription());		
		vf.startEdit(VFrame.NODETYPEDESCRIPTION,selectedNodeType);		
	}
}


//-------------------------------------------------------------------------------
//	                        Edit Actions for Slots 
//-------------------------------------------------------------------------------


/**
 * clear everything from the drawing  area and draw a slot; 
 * then offer the option to edit the slot itself (rather than its values)
 * @param s
 */
protected void drawSlot(Slot s, boolean offerEdit)
{	
	drawArea.drawSlot(s);
	selectedSlot = s;
	selectedNodeType = null;
	selectedSlotValue = null;
	
	// allow the user to edit the slot, including changing the slot name
	if ((offerEdit) && (Util.getYesNo("Edit slot '" + s.getName() + "' (not its values)?", vf))) editSlot(true);
	
	// e.g. to enable the menu item to delete the slot, and disable deleting an old slot value or node type
	refreshEditModelMenus(false,null);

}

/**
 * 
 */
private void addSlot()
{
	setEditState(MODEL);
	
	String slotName = getNewSlotName("");
	
	// valid new slot name - not cancelled
	if (slotName != null) try
	{
		Slot slot = new Slot(model(),slotName,"", "");
		model().getSlots().put(slotName, slot);
		selectedSlot = slot;
		
		drawArea.drawSlot(slot);
		
		modelChangedSinceSaved = true;
		refreshEditModelMenus(false,null);
		
		editSlot(false);
	}
	catch (Exception ex ) {}
}

/**
 * remove the selected slot from the model
 */
private void removeSlot()
{
	if (selectedSlot != null)
	{
		model().getSlots().remove(selectedSlot.getName());
		
		modelChangedSinceSaved = true;
		refreshEditModelMenus(false,null);
	}
}

/**
 * 
 * @param oldSlotName
 * @return a new slot name, or null if the user gives up
 */
private String getNewSlotName(String oldSlotName)
{
	String slotName = oldSlotName;
	boolean isOKSlotName = false;
	String title = "Slot name";
	
	// get a new slot name from the user - or null if cancelled
	while ((!isOKSlotName) &&  (slotName != null))
	{
		slotName = Util.getShortUserInput(title, vf, slotName);
		if (slotName != null)
		{
			// the slot name must be new or unchanged
			isOKSlotName = ((model().getSlot(slotName) == null)||(slotName.equals(oldSlotName)));
			// only used if the slot name is not OK
			title = "Slot Name '" + slotName + "' is already used";
		}
	}
	
	return slotName;
	
}

/**
 * Edit a slot (not its values; you need to select a value to do that)
 * @param changeNameallow name changes if true
 */
private void editSlot(boolean changeName)
{
	// show the description of the slot in the text area
	act.showSlot(selectedSlot);

	String newName = null;
	String oldName = selectedSlot.getName();
	if (changeName) newName = getNewSlotName(oldName);
	if (newName != null) selectedSlot.setName(newName);
	
	// get an allowed set of node types, or null
	String title = "Node types this slot can appear on";
	String allowedNodes = getAllowedNodeTypes(title,selectedSlot.getAllowedNodeTypeNames());
	
	// change the nodes the slot can appear on
	if (allowedNodes != null) try
	{
		selectedSlot.setAllowedTypes(allowedNodes);
	}
	catch (Exception ex) {}
	
	// enable the changes to be saved
	modelChangedSinceSaved = true;
	refreshEditModelMenus(false,null);
	
	if (Util.getYesNo("Edit description of slot?",vf))
	{
		String bareDescription = selectedSlot.getDescription();
		act.explain(bareDescription);	
		saveItem.setEnabled(true);
		
		vf.startEdit(VFrame.SLOTDESCRIPTION,selectedSlot);		
	}
	
	if (newName != null) 
	{
		// re-display, in drawing areas; but do not offer another edit
		drawSlot(selectedSlot,false);
		// act.showSlot(selectedSlot);
		
		// make all necessary changes in the language model - change the slot name wherever it occurs		
		changeSlotName(oldName, newName);
	}
	
}

/**
 * find all occurrences of an old slot name in the language model, 
 * and change them to a new slot name, with the same value
 * 
 * @param oldName
 * @param newName
 */
private void changeSlotName(String oldName, String newName)
{
	// words
	for (Enumeration<FSNode> en = model().getWords().elements(); en.hasMoreElements();)
		en.nextElement().changeSlotName(oldName, newName);
			
	// sentences
	for (Enumeration<FSNode> en = model().getSentences().elements(); en.hasMoreElements();)
		en.nextElement().changeSlotName(oldName, newName);
	
	// words to learn and learning examples
	for (Enumeration<LearningStage> en = model().getLearningStages().elements(); en.hasMoreElements();)
	{
		LearningStage stage = en.nextElement();
		for (WordToLearn wl: stage.getLearnWords())
		{ 
			FSNode toLearn = wl.getWordFS();
			if (toLearn != null) toLearn.changeSlotName(oldName, newName);
			
			for (FSNode fs: wl.getExamples()) fs.changeSlotName(oldName, newName);
		}
	}
	
	vf.getEditLanguage().changedSinceSaved = true;
			
}

//-------------------------------------------------------------------------------
//	                            Edit actions for Slot values 
//-------------------------------------------------------------------------------




/**
 * add a new SlotValue, under an existing SlotValue, or direct on a slot
 * @param slotVal
 */
private void addValueTo()
{
	Slot slot = selectedSlotValue.getSlot();
	boolean addDirectToSlot = selectedSlotValue.representsSlot();
	SlotValue parent = selectedSlotValue;
	if (addDirectToSlot) parent = null;
	
	String value = Util.getShortUserInput("Slot value", vf, "");
	// "" means no description yet
	SlotValue newValue = new SlotValue(value,"", slot, parent);
	
	// add the new value to the selected parent value or direct on the slot
	if (addDirectToSlot) slot.getValues().add(newValue);
	else parent.getSubValues().add(newValue);
	
	drawArea.drawSlot(slot);
	setEditState(MODEL);
	modelChangedSinceSaved = true;
	refreshEditModelMenus(false,null);
	
	editSlotValueDescription(newValue);
	
}

/**
 * remove the selected slot value (and all its descendant values)
 */
private void removeValue()
{
	if (selectedSlotValue !=  null)
	{
		SlotValue parent = selectedSlotValue.getParent();
		Slot ownerSlot = selectedSlotValue.getSlot();
		
		// remove the slot value either from its parent slot value, or from the owning slot (if it is a top-level value)
		if (parent != null) parent.getSubValues().remove(selectedSlotValue);
		else ownerSlot.getAllValues().remove(selectedSlotValue);
		
		
		drawArea.drawSlot(ownerSlot);
		setEditState(MODEL);
		modelChangedSinceSaved = true;
		refreshEditModelMenus(false,null);

	}
}


/**
 * edit a selected slot value - changing either the value or the description or both
 */
private void editSlotValue()
{
	String oldValue = selectedSlotValue.getValue();
	String newValue = Util.getShortUserInput("New value", vf, oldValue);
	if (newValue == null) newValue = oldValue;
	if (!newValue.equals(oldValue))
		selectedSlotValue.setValue(newValue);
	
	Slot slot = selectedSlotValue.getSlot();
	
	drawArea.drawSlot(slot);
	setEditState(MODEL);
	modelChangedSinceSaved = true;
	refreshEditModelMenus(false,null);
	
	editSlotValueDescription(selectedSlotValue);
	
	
}

/**
 * 
 * @param sv
 */
private void editSlotValueDescription(SlotValue sv)
{
	if (Util.getYesNo("Edit description of slot value?",vf))
	{
		act.explain(sv.getDescription());		
		vf.startEdit(VFrame.SLOTVALUEDESCRIPTION,sv);		
	}
	
}



//-------------------------------------------------------------------------------
//	               Getting a valid set of node types 
//        (for child node types of a node, or node types a slot can be on) 
//-------------------------------------------------------------------------------

/**
 * 
 * @return a non-empty set of valid node type names entered by the user
 * or null if cancelled
 */
private String getAllowedNodeTypes(String firstTitle,String previous)
{
	String allowedNodes = "";
	boolean nodesOK = false;
	
	while ((!nodesOK) && (allowedNodes != null))
	{
		allowedNodes = Util.getShortUserInput(firstTitle, vf, previous);
		if (allowedNodes != null)
		{
			String bNodes = bannedNodes(allowedNodes);
			nodesOK = (bNodes.length() == 0);
			firstTitle = "Node types " + bNodes + "are not valid";
		}
	}
	
	if ((allowedNodes != null) && (allowedNodes.length() == 0)) allowedNodes = null;
	return allowedNodes;
	
}

/**
 * names of user entered node types that are not allowed
 * @param nodes
 * @return
 */
private String bannedNodes(String nodes)
{
	String bNodes = "";
	StringTokenizer st = new StringTokenizer(nodes," ");
	
	while (st.hasMoreTokens())
	{
		String node = st.nextToken();
		if (model().getNodeType(node) == null) 
			bNodes = bNodes + node + " ";
	}
	
	return bNodes;
}

//-------------------------------------------------------------------------------
//	                          Other Actions
//-------------------------------------------------------------------------------

private void editTopic()
{
	if (selectedTopic != null)
	{
		Util.message("Editing topic " + selectedTopic.getName());
		
		// show the topic without its title
		act.showTopic(selectedTopic.getId(), false);
		
		// enable editing the text
		vf.startEdit(VFrame.TOPIC,selectedTopic);		
	}
}



/**
 * validate the full language model and report any errors
 */
private void checkLanguage()
{
	try
	{
		model().validateLanguageModel();
		model().writeValidationErrors();		
	}
	catch (Exception ex) {}
}


/**
 * save the model definition and the user guide, overwriting the previous file (make a backup first!)
 * 
 * If model changes (changes to slot names) caused language changes, save them as well
 */
private void doSave() 
{
	try
	{
		model().saveModelAndGuide();
		modelChangedSinceSaved = false;
		refreshEditModelMenus(false,null);
		
		if (vf.getEditLanguage().changedSinceSaved)
		{
			model().saveLanguageFiles();
			vf.getEditLanguage().changedSinceSaved = false;
		}
	}
	catch (Exception ex) {act.explain(ex.getMessage());}
}


private void doFinish() 
{
	drawArea.clear();
	setEditState(EditLanguage.NONE);
}





}
