package displayPack;

import java.util.Hashtable;
import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import modelPack.LanguageModel;
import utilPack.Util;

public class Topic {
	
	private LanguageModel model;
	
	private String topicName;
	public String getName() {return topicName;}
	
	private Vector<String> paragraphs = new Vector<String>();
	public Vector<String> getParas() {return paragraphs;}
	
	private Vector<Topic> subTopics = new Vector<Topic>();
	public Vector<Topic> getSubTopics() {return subTopics;}
	
	private String id;
	public String getId() {return id;}
	
	public Topic getParent() {return parent;}
	private Topic parent;
	
	//--------------------------------------------------------------------------
	//                                Constructor
	//--------------------------------------------------------------------------
	
	/**
	 * A Topic could in principle have both sub-topics and paragraphs;
	 * but actual topics will have only one or the other.
	 * @param topicEl
	 */
	public Topic(LanguageModel model,Element topicEl, Topic parent)
	{
		this.model = model;
		this.parent = parent;
		
		// non-repeating id
		id = model.getId();
		
		topicName = topicEl.getAttribute("name");
		
		for (Element subEl: Util.namedChildElements(topicEl, "topic")) subTopics.add(new Topic(model,subEl,this));
		
		for (Element paraEl: Util.namedChildElements(topicEl, "p")) paragraphs.add(Util.getText(paraEl));
	}
	
	//--------------------------------------------------------------------------
	//                             XML form to save
	//--------------------------------------------------------------------------
	
	
	/**
	 * 
	 * @param doc
	 * @return
	 */
	public Element xmlForm(Document doc) throws Exception
	{
		Element topicEl = Util.newElement(doc, "topic");
		topicEl.setAttribute("name", topicName);
		
		for (String para: paragraphs) 
		{
			Element paraEl = Util.textElement(doc, "p", para);
			topicEl.appendChild(paraEl);
		}
		
		for (Topic subTop: subTopics)
			topicEl.appendChild(subTop.xmlForm(doc));
		
		
		return topicEl;
	}
	
	
	//--------------------------------------------------------------------------
	//                             Various methods
	//--------------------------------------------------------------------------
	
	/**
	 * 
	 * @param leafTopics
	 */
	public void storeTopicIds(Hashtable<String,Topic> leafTopics)
	{
		if (subTopics.size() == 0)
			leafTopics.put(getId(), this);
		else
			for (Topic sub: subTopics) sub.storeTopicIds(leafTopics);
	}
	
	/**
	 * 
	 * @return title showing the trail of topics and sub-topics down to this one
	 */
	public String getTitle()
	{
		if (parent == null) return getName().toUpperCase();
		else return parent.getTitle() + " > " + getName().toUpperCase();		
	}
	
	/**
	 * 
	 * @return
	 */
	public String getCompleteContent(boolean withTitle)
	{
		String content = "";
		if (withTitle) content = getTitle() + "\n\n";
		
		for (String para:getParas()) content = content + para + "\n\n";
		
		return content;
	}
	
	/**
	 * update the paragraphs to reflect the result of editing
	 * @param editResult
	 */
	public void setParagraphs(String result)
	{
		// writeTrace(result);
		
		// re-zero paragraphs
		paragraphs = new Vector<String>();
		String para = "";
		
		for (int c = 0; c < result.length(); c++)
		{
			char ch = result.charAt(c);
			char[] ch1 = new char[1];
			ch1[0] = ch;
			int i = ch;
			
			// new line marker
			if (i == 10)
			{
				// if there are two or more '10' newlines in succession, do not add empty paragraphs for them
				if (para.length() > 0) paragraphs.add(para);
				para = ""; // start building the next paragraph
			}
			// character to go in a paragraph
			else para = para + new String(ch1); 
		}
		// add final paragraph if there are no newlines after it
		if (para.length() > 0) paragraphs.add(para);		
	}
	
	

}
