package displayPack;

import java.awt.Adjustable;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

import generatorPack.ExGenerator;
import modelPack.FSNode;
import modelPack.LanguageModel;
import modelPack.LearningStage;
import modelPack.NodeType;
import modelPack.Slot;
import modelPack.SlotValue;
import modelPack.WordToLearn;
import utilPack.Util;

public class VFrame extends JFrame {
	
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	
	// size index = 0 for small window, 1 for large window
	public void setSizeIndex(int sizeIndex) {this.sizeIndex = sizeIndex;}
	// use the large window size at first
	private int sizeIndex = 1;
		
	// top corner of window on screen
	private int left = 50;
	private int top = 50;
	
	// total window size, in small and large configurations
	private int[] heights = {700,1300};
	private int[] widths = {1200,1800};
	
	public int getDrawAreaWidth() {return drawAreaWidth;}
	int drawAreaWidth;
	
	int drawingAreaHeight;
	
	// width of left-hand button area
	private int buttonAreaWidth = 160;
		
	// depth of lower explanation area
	private int[] explanationHeights = {200,500};
	
	// depth allowed for menu bar
	private int menuBarHeight = 10;
	
	// general width insets
	private int[] widthInsets = {30,100};
	
	private int heightInset = 30;

	// top area for drawing
	public DrawArea getDrawArea() {return drawArea;} // extends Canvas
	private DrawArea drawArea = new DrawArea(this);

	// lower area for text explanation
	public JTextArea explanationArea = new JTextArea(); 
	JScrollPane explanationPane = new JScrollPane(explanationArea); // explanation area scrolls vertically
	
	// tree of menus to send in XML form to the client
	private MenuElement menus = new MenuElement("topMenu");
	public MenuElement getMenus() {return menus;}
	

	  //------------------------------------------------------------------------------------------------
	  //                             Non-graphic variables
	  //------------------------------------------------------------------------------------------------
	
	private Actions act;
	public Actions getAct() {return act;}
	
	EditLanguage editLanguage;
	public EditLanguage getEditLanguage() {return editLanguage;}
		
	EditModel editModel;
	public EditModel getEditModel() {return editModel;}

	// a language model
	private LanguageModel model;
	public void setModel(LanguageModel model) {this.model = model;}
	public LanguageModel getModel() {return model;}
	
	// for alphabetically sorted menus
	public static String[]  alphabet = {"A","B","C","D","E","F","G","H","I","J","K","L","M","N","O","P","Q","R","S","T","U","V","W","X","Y","Z"};
	
	private ExGenerator generator;

	  //------------------------------------------------------------------------------------------------
	  //                              constructor and startup
	  //------------------------------------------------------------------------------------------------
	
	
	/**
	 * 
	 */
	public VFrame(boolean show)
	{
		super("Demonstration of Language Learning");

		// stop the whole application on close
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

		// these two lines fixed the problem of components like buttons disappearing, for some unknown reason.
		getContentPane().setLayout(null);
	    setResizable(false);
	    
	    // main actions for learning, understanding, generating
	    act = new Actions(this,drawArea);
	    
		// for editing a language model instance
		editLanguage = new EditLanguage(this,drawArea);
	    
		// for editing the model definition
		editModel = new EditModel(this,drawArea);
		
		// set the coordinates of everything
		frameSetup(show);
		
		
		
	}
	
	/**
	 * 
	 */
	public void reSize()
	{
	    // coordinates of the whole frame on the screen
	    setBounds(left, top, widths[sizeIndex], heights[sizeIndex]);
	    
	    // drawing area
	    drawAreaWidth = widths[sizeIndex] - buttonAreaWidth - widthInsets[sizeIndex];
	    drawingAreaHeight = heights[sizeIndex] - explanationHeights[sizeIndex] - menuBarHeight;
	    drawArea.setBounds(buttonAreaWidth,menuBarHeight, drawAreaWidth, drawingAreaHeight);
	    
	    // explanation pane
	    explanationPane.setBounds(buttonAreaWidth, heights[sizeIndex] - explanationHeights[sizeIndex], drawAreaWidth, explanationHeights[sizeIndex] - menuBarHeight - 2*heightInset);
	    
	    // redisplay
	    validate();
		
	}
	
	/**
	 * 
	 */
	public void frameSetup(boolean show)
	{
	    // coordinates of the whole frame on the screen
	    setBounds(left, top, widths[sizeIndex], heights[sizeIndex]);
	    
	    drawAreaWidth = widths[sizeIndex] - buttonAreaWidth - widthInsets[sizeIndex];
	    
	    {
	    	// set up the drawing area
	    	drawingAreaHeight = heights[sizeIndex] - explanationHeights[sizeIndex] - menuBarHeight;
		    drawArea.setBounds(buttonAreaWidth,menuBarHeight,drawAreaWidth, drawingAreaHeight);
		    
		    // horizontal scrollbar
		    JScrollBar horizontalScrollBar = new JScrollBar(Adjustable.HORIZONTAL);
		    horizontalScrollBar.setBounds(buttonAreaWidth, drawingAreaHeight - 30, drawAreaWidth - 30, 30);
		    AdjustmentListener adjustmentListener = new AdjustmentListener() {
		        public void adjustmentValueChanged(AdjustmentEvent adjustmentEvent) {
		        	drawArea.setHorizontalScrollValue(adjustmentEvent.getValue());
		        	drawArea.repaint();
		        }
		      };
		    horizontalScrollBar.addAdjustmentListener(adjustmentListener);		    
		    add(horizontalScrollBar);
		    
		    // vertical scrollbar
		    JScrollBar verticalScrollBar = new JScrollBar(Adjustable.VERTICAL);
		    verticalScrollBar.setBounds(buttonAreaWidth + drawAreaWidth - 30, menuBarHeight, 30, drawingAreaHeight);
		    AdjustmentListener verticalAdjustmentListener = new AdjustmentListener() {
		        public void adjustmentValueChanged(AdjustmentEvent adjustmentEvent) {
		        	drawArea.setVerticalScrollValue(adjustmentEvent.getValue());
		        	drawArea.repaint();
		        }
		      };
		    verticalScrollBar.addAdjustmentListener(verticalAdjustmentListener);		    
		    add(verticalScrollBar);
		    
		    drawArea.addVertScrollBar(verticalScrollBar);
		    
		    getContentPane().add(drawArea);
	    }
	    
	    	    
	    // explanation pane and area
	    explanationArea.setLineWrap(true);
	    explanationArea.setWrapStyleWord(true);
	    explanationArea.setText("");
	    explanationArea.setEditable(false);
	    explanationArea.setFont(new Font(Font.DIALOG,Font.PLAIN,20));
	    explanationPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
	    explanationPane.setBounds(buttonAreaWidth, heights[sizeIndex] - explanationHeights[sizeIndex], drawAreaWidth, explanationHeights[sizeIndex] - menuBarHeight - 2*heightInset);
	    explanationArea.setVisible(true);
	       
	    getContentPane().add(explanationPane);
	    
	    // menus
	    makeInitialMenus();
	    	    
	    // buttons
	    makeButtons();
	    
	    // display
	    if (show)
	    {
		    setVisible(true);
		    validate();
	    }

	}
	
	/**
	 * called every time the vertical scrolling is moved
	 * @param adjustValue
	 */
	public void scrollAdjust(int adjustValue)
	{
		drawArea.repaint();
		validate();
	}
	
	
	//-------------------------------------------------------------------------------
	//                    Populate menus from model data
	//-------------------------------------------------------------------------------
	
	// Menu bar and menus
	JMenuBar qMBar = new JMenuBar();
	JMenu guideMenu = new JMenu("Guide");
	JMenu wordMenu = new JMenu("Words");
	JMenu sentenceMenu = new JMenu("Sentences");
	JMenu learnMenu = new JMenu("Learn");	
	JMenu editLanguageMenu = new JMenu("Edit Language");
	JMenu editModelMenu = new JMenu("Edit Model");
	
	/** 
	 * after any automatic learning, the Learning menu and WordToLearn objects are changed in class Learner, 
	 * so editing and saving is entirely disabled - to stop saving the automatically learnt words 
	 * */
	public JMenu getLearnMenu() {return learnMenu;}
	public void disableAllEdits()
	{
		editLanguageMenu.setEnabled(false);
		editModelMenu.setEnabled(false);
	}


	/**
	 * Make as much of the menus as can be made before the model file is read
	 */
	private void makeInitialMenus()
	  {
			// All menus are to be populated entirely depending on model data
	        qMBar.add(guideMenu);
		    qMBar.add(wordMenu);
		    qMBar.add(sentenceMenu);
		    qMBar.add(learnMenu);
		    
		    // edit language menu h
		    qMBar.add(editLanguageMenu);
		    editLanguage.setUpEditMenu(editLanguageMenu);
		    
		    // edit model menu 
		    qMBar.add(editModelMenu);
		    editModel.setUpEditMenu(editModelMenu);

		    setJMenuBar(qMBar);		    
	  }
	

	/**
	 * fill the menus when the model is first loaded
	 * @param model
	 */
	public void fillMenus(LanguageModel model)
	{
		// menu tree to send to the client
		MenuElement topicElement = new MenuElement("Help");
		MenuElement wordElement = new MenuElement("Words");
		MenuElement sentenceElement = new MenuElement("Sentences");
		MenuElement learnElement = new MenuElement("Learning");
		MenuElement modelElement = new MenuElement("Model");
		menus.addChild(topicElement);
		menus.addChild(wordElement);
		menus.addChild(sentenceElement);
		menus.addChild(learnElement);
		menus.addChild(modelElement);
		
		// make the user guide topic menu
		Vector<Topic> allTopics = model.getTopics();		
		for (Topic topic: allTopics) addTopic(topic, guideMenu,topicElement);
		
		// populate the word menu. False = not for editing
		addFeatureStructureItems(model.getWords(), wordMenu,act, EditLanguage.WORD, EditLanguage.SHOW,wordElement);

		// populate the sentences menu. False = not for editing
		addFeatureStructureItems(model.getSentences(), sentenceMenu,act, EditLanguage.SENTENCE, EditLanguage.SHOW,sentenceElement);
		
		// populate the learn menu
		addLearnItems(model.getLearningStages(),learnMenu,learnElement);
		
		// set up the language edit menus, with nothing selected
		editLanguage.refreshEditLanguageMenus(null,null,false);
		
		// set up the model edit menus, with nothing selected;  true = atStart
		editModel.refreshEditModelMenus(true,modelElement);
		
		// diagnostic write of nested menu structures to be sent to the client
		// menus.write("");
}
	
	
	/**
	 * recursive addition of topics
	 * @param topic
	 * @param menu
	 */
	private void addTopic(Topic topic, JMenu menu, MenuElement topicElement)
	{
		
		// topic has sub-topics. Make it a menu and add them
		if (topic.getSubTopics().size() > 0)
		{
			JMenu topicMenu = new JMenu(topic.getName());
			menu.add(topicMenu);

			// for the web client
			MenuElement subElement = new MenuElement(topic.getName());
			topicElement.addChild(subElement);

			// recursion to sub-topics
			for (Topic subTopic: topic.getSubTopics()) addTopic(subTopic, topicMenu,subElement);
		}
		
		// topic has no sub-topics. Make it a menu item
		else
		{
			JMenuItem topicItem = new JMenuItem(topic.getName());
			final String id = topic.getId();
		    topicItem.addActionListener(new ActionListener() {
		    	final String topicId = id;
		        public void actionPerformed(ActionEvent ev) {
		        	act.showTopic(topicId,true);
		        }
		    });
			menu.add(topicItem);
			
			// for the web client
			MenuElement subElement = new MenuElement(topic.getName());
			topicElement.addChild(subElement);
		}
	}
	
	
	/**
	 * 
	 * @param words
	 */
	public static void addFeatureStructureItems(Hashtable<String,FSNode> words, JMenu addTo, Actions act, int type, int actionType,MenuElement wordElement)
	{
		addTo.removeAll();
		
		// Hashtable of all words or sentences, keyed by initial capital letter
		Hashtable<String,Vector<String>> names = new Hashtable<String,Vector<String>>();
		
		for (Enumeration<FSNode> en = words.elements(); en.hasMoreElements();)
		{
			FSNode word = en.nextElement();
			String wName = word.getName();
			
			// upper case initial letter, excluding "'"
			String initial = wName.substring(0,1).toUpperCase();
			if (initial.equals("'")) initial = wName.substring(1,2).toUpperCase();
			
			Vector<String> nl = names.get(initial);
			if (nl == null) nl = new Vector<String>();
			nl.add(wName);
			names.put(initial,nl);
		}
		
		for (int i = 0; i < alphabet.length; i++)
		{
			String initial = alphabet[i];
			Vector<String> nl = names.get(initial);
			if (nl != null)
			{
				JMenu initMenu = new JMenu(initial);
				addTo.add(initMenu);

				// for the web client
				MenuElement initElement = new MenuElement(initial);
				wordElement.addChild(initElement);

				for (String w: nl)
				{
					JMenuItem wItem = new JMenuItem(w);
				    wItem.addActionListener(new ActionListener() {
				        public void actionPerformed(ActionEvent ev) {
				        	String value = ev.getActionCommand();
				        	act.showFS(value,type,actionType);
				        }
				    });
					initMenu.add(wItem);

					// for the web client
					MenuElement wElement = new MenuElement(w);
					initElement.addChild(wElement);
				}
			}
		}
		
		// to make the alpha sub-menus look wider
		String widener = "---------";
		JMenu wideMenu = new JMenu(widener);
		addTo.add(wideMenu);
	}
	
	/**
	 * 
	 * @param stages
	 * @param lMenu
	 */
	public void addLearnItems(Vector<LearningStage> stages, JMenu lMenu, MenuElement learnElement)
	{
		
		for (LearningStage stage: stages)
		{
			String stageName = stage.getName();
			JMenu stageMenu = new JMenu(stageName);
			lMenu.add(stageMenu);
			
			// for the web client
			MenuElement stageElement = new MenuElement(stageName);
			if (learnElement != null) learnElement.addChild(stageElement);
			
			for (WordToLearn wl: stage.getLearnWords())
			{
				String wName = wl.getWordName();

				// add a menu item for each word to be learned
				JMenuItem lItem = new JMenuItem(wName);
				stageMenu.add(lItem);			
			    lItem.addActionListener(new ActionListener() {
			        public void actionPerformed(ActionEvent ev) {
			        	act.startLearn(ev,stageName);
			        }
			    });

			    // for the web client
			    MenuElement wlElement = new MenuElement(wName);
				stageElement.addChild(wlElement);
			}
		}
	}
	//-------------------------------------------------------------------------------
	//                        Make Buttons
	//-------------------------------------------------------------------------------
	
	
	// Buttons
	JButton understandButton = new JButton("Understand");
	JButton generateButton = new JButton("Generate");
	JButton stepButton = new JButton("Step");
	JButton showHideButton = new JButton("Show");
	JButton acceptEditButton = new JButton("Accept Edit");
	JButton cancelEditButton = new JButton("Cancel Edit");
	JButton autoLearnButton = new JButton("Auto-Learn");
	JButton saveClassButton = new JButton("Save Class");
	
	int buttonTop = 30;
	int buttonLeft = 30;
	int buttonSpacing = 60;
	int buttonWidth = 100;
	int buttonHeight = 30;
	
	/**
	 * 
	 */
	private void makeButtons()
	{		
		JButton[] buttons = new JButton[8];
		buttons[0] = understandButton;
		buttons[1] = generateButton;
		buttons[2] = stepButton;
		buttons[3] = showHideButton;
		buttons[4] = acceptEditButton;
		buttons[5] = cancelEditButton;
		buttons[6] = autoLearnButton;
		buttons[7] = saveClassButton;
		
		for (int b = 0; b < buttons.length; b++)
		{
			buttons[b].setSize(buttonWidth, buttonHeight);
			buttons[b].setBounds(buttonLeft, buttonTop + b*buttonSpacing, buttonWidth, buttonHeight);
			final int bb = b;
			
			buttons[b].addActionListener(new ActionListener() {
		    	final int buttonId = bb;
		        public void actionPerformed(ActionEvent ev) {
		        	act.doButtonAction(buttonId);
		        }
		    });
			
			getContentPane().add(buttons[b]);
			buttons[b].setVisible(true);
			buttons[b].setEnabled(false);
		}
		
		// auto-learning and saving the resulting class are always enabled
		autoLearnButton.setEnabled(true);
		saveClassButton.setEnabled(true);
	}
	   

	  //------------------------------------------------------------------------------------------------
	  //                            Editing text in the explanation area
	  //------------------------------------------------------------------------------------------------
	
	// the text of a description after it has been edited
	private String editResult = null;
	public String getEditResult() {return editResult;}
	
	// the object whose description is currently being edited
	private Object editedThing;
	
	// the type of the object whose description is being edited
	private int editedType;
	
	// types of object whose text description can be edited
	public static int NODETYPEDESCRIPTION = 0;
	public static int SLOTDESCRIPTION = 1;
	public static int SLOTVALUEDESCRIPTION = 2;
	public static int FSDESCRIPTION = 3;
	public static int TOPIC = 4;
	
	/**
	 * start editing the text description of an object of some type
	 * @param type the integer object type 0..3
	 * @param thing the object whose description is being edited
	 */
	public void startEdit(int type, Object thing)
	{
		editedType = type;
		editedThing = thing;
		
		acceptEditButton.setEnabled(true);
		cancelEditButton.setEnabled(true);
		explanationArea.setEditable(true);		
	}
	
	/**
	 * cancel an edit of any description
	 */
	public void doCancelEdit()
	{
		act.explain("Edit cancelled");
		finishEdit();
	}

	
	/**
	 * accept an edit of a description
	 */
	public void doAcceptEdit()
	{
		editResult = explanationArea.getText();
		
		if (editedType == NODETYPEDESCRIPTION)
		{
			NodeType nt  = (NodeType)editedThing;
			nt.setDescription(editResult);
			model.addNodeType(nt);
			act.explain(nt.getExplanation());
		}
		
		else if (editedType == SLOTDESCRIPTION)
		{
			Slot sl = (Slot)editedThing;
			sl.setDescription(editResult);
			act.explain(sl.getExplanation());
		}
		
		else if (editedType == SLOTVALUEDESCRIPTION)
		{
			SlotValue sv = (SlotValue)editedThing;
			sv.setDescription(editResult);
			act.explain(editResult); // slot values do not have an explanation more than their description
		}
		
		else if (editedType == FSDESCRIPTION)
		{
			FSNode fs = (FSNode)editedThing;
			fs.setDescription(editResult);
			act.explain(editResult); 
		}
		
		else if (editedType == TOPIC)
		{
			Topic top = (Topic)editedThing;
			
			// reset the topic paragraphs to the result of editing
			top.setParagraphs(editResult);
			
			// re-show the topic with its title
			act.showTopic(top.getId(), true);
			
			// enable saving this edit
			editModel.setChangedSinceSaved(true);
			editModel.refreshEditModelMenus(false,null);
		}

		finishEdit();
	}
	
	/**
	 * finish an edit of a description or topic, after accepting or cancelling
	 */
	public void finishEdit()
	{
		editResult = null;
		editedThing = null;
		
		acceptEditButton.setEnabled(false);
		cancelEditButton.setEnabled(false);
		explanationArea.setEditable(false);		
	}
	
	  //------------------------------------------------------------------------------------------------
	  //                              Unsupervised learning
	  //------------------------------------------------------------------------------------------------
	
	/**
	 * on pressing the 'Auto-Learn' button,
	 * do one or more cycles of learning
	 */
	public void doAutoLearn()
	{
		// zeroth cycle is done once only
		if (generator.getCycleNo() == 0) try {generator.oneCycle();} catch(Exception ex) {}
		
		// do all other cycles one or more times for each button press
		else for (int c = 0; c < generator.cyclesPerPress;c++)
		{
			try {generator.oneCycle();}
			catch(Exception ex) { Util.message("Error making examples or learning: " + ex.getMessage()); ex.printStackTrace();}
		}
	}

	
	/**
	 * on pressing the 'Save Class' button
	 */
	public void doSaveClass()
	{
		try {generator.saveLearntClass();}
		catch(Exception ex) 
		{
			Util.message("Error saving generated class: " + ex.getMessage()); ex.printStackTrace();
		}
	}

	
	/**
	 * on startup
	 * @param generator
	 */
	public void setExGenerator(ExGenerator generator)
	{
		this.generator= generator;
	}
	
	  //------------------------------------------------------------------------------------------------
	  //                              bits & bobs
	  //------------------------------------------------------------------------------------------------
	
	/**
	 * temporary
	 * @param s
	 */
	public void frameMessage(String s) {Util.message(s);}
	


}
