package generatorPack;

import java.util.Vector;

import org.w3c.dom.Element;

import utilPack.Util;

/**
 * A category used to generate learning examples.
 * motivated by and loosely based on Categorial grammars - 
 * but this class is not part of the language model, 
 * and is only used to generate utterances as learning examples
 *
 */

public class Category {
	
	public String getName() {return name;}
	private String name;
	
	Vector<Production> productions = new Vector<Production>();
	
	private double allWeights = 0.0;
		
	
	//--------------------------------------------------------------------------
	//	                        Constructor
	//--------------------------------------------------------------------------
	
	public Category(Element catRoot, ExGenerator generator)
	{
		name = catRoot.getAttribute("name");
		String description = catRoot.getAttribute("description"); // not used
		
		for (Element prodRoot: Util.namedChildElements(catRoot, "made"))
		{
			Production prod = new Production(prodRoot,this,generator);
			productions.add(prod);
			allWeights = allWeights + prod.getWeight();
		}
	}

	
	//--------------------------------------------------------------------------
	//	                        Methods
	//--------------------------------------------------------------------------
	
	/**
	 * check all productions
	 * and return a Vector of real words in the category
	 * @throws Exception
	 */
	public Vector<String> check() throws Exception
	{
		Vector<String> realWords = new Vector<String>();
		for (Production prod: productions) 
		{
			Vector<String>  prodWords = prod.check();
			for (String pw: prodWords) realWords.add(pw);
		}
		return realWords;
	}
	
	/**
	 * get a random production, with probability depending on their weights
	 * @return
	 */
	public Production getRandomProduction()
	{
		double rand = Math.random(); // a random number in 0...1 - delta (it should never quite get to 1)
		double topWeight = allWeights*rand; // a random number in 0...allWeights - delta
		double weight = 0;
		
		for (Production prod: productions)
		{
			weight = weight + (double)prod.getWeight(); // after the last production, weight = allWeights
			if (weight > topWeight) return prod;
		}
		return null;
	}

}
