package generatorPack;

import java.util.StringTokenizer;
import java.util.Vector;

import org.w3c.dom.Element;

import modelPack.FSNode;
import utilPack.Util;

/**
 * A production used to generate learning examples.
 * motivated by and loosely based on various grammars - 
 * but this class is not part of the language model, 
 * and is only used to generate utterances as learning examples
 *
 */


public class Production {
	
	private Category cat;
	
	private ExGenerator generator;
	
	public int getWeight() {return weight;}
	private int weight;
	
	private Vector<String> catNames = new Vector<String>();
	public Vector<String> getCatNames() {return catNames;}
	
	// short names of words - not spaced names
	private Vector<String> words = new Vector<String>();
	public Vector<String> getWords() {return words;}
	
	public boolean isLeaf() {return words.size() > 0;}
		
	
	//--------------------------------------------------------------------------
	//	                        Constructor
	//--------------------------------------------------------------------------
	
	public Production(Element prodRoot, Category cat, ExGenerator generator)
	{
		this.cat = cat;
		this.generator = generator;
		
		String w = prodRoot.getAttribute("weight");
		weight = Integer.parseInt(w);
		
		String nn = prodRoot.getAttribute("cats");
		if (nn != null)
		{
			StringTokenizer st = new StringTokenizer(nn," ");		
			while (st.hasMoreTokens()) catNames.add(st.nextToken());
		}
		
		String ww = prodRoot.getAttribute("words");
		if (ww != null)
		{
			StringTokenizer st = new StringTokenizer(ww," ");		
			while (st.hasMoreTokens()) words.add(st.nextToken());
		}
	}
	
	//--------------------------------------------------------------------------
	//	                        Methods
	//--------------------------------------------------------------------------
	
	
	/**
	 * 
	 * @return
	 * @throws Exception
	 */
	public Vector<Category> getCategories() throws Exception
	{
		Vector<Category> cats = new Vector<Category>();
		for (String catName: catNames)
		{
			Category cat = generator.getCategory(catName);
			if (cat == null) throw new Exception("There is no category " + catName);
			cats.add(cat);
		}
		return cats;
	}
	
	/**
	 * check that the categories or words in this production exist; 
	 *  return a Vector of real words in the production
	 */
	public Vector<String>  check() throws Exception
	{
		Vector<String> realWords = new Vector<String>();
		getCategories(); // throws exception if any categories are undefined
		String missingWords = "";
		
		for (String word:words)
		{
			FSNode w = generator.getModel().getWords().get(word);
			if (w == null) missingWords = missingWords + word + " ";
			else 
			{
				String spacedName  = w.getSpacedName(); // name in spaced form r-u-n etc.
				String shortName = w.getName();
				realWords.add(spacedName);

				// store the FSNode and short name for each real word
				generator.realWordFS.put(spacedName, w);
				generator.shortToSpaced.put(shortName, spacedName);
				
				// not any words that create a nested context, and are problematic for learning other words
				if (w.makesNestedContext())
					Util.message("word " + word + " creates a nested context.");
			}
		}
		
		if (missingWords.length() > 0) throw new Exception("Missing words: " + missingWords + " in a production in category " + cat.getName());
		
		return realWords;
	}
	
	/**
	 * get a random word, with equal probability for all words
	 */
	public String getRandomWord() throws Exception
	{
		if (!isLeaf()) throw new Exception("Production in category " + cat.getName() + " is not a leaf production");
		
		double rand = Math.random(); // a random number in 0 .. 1 - delta
		double topWeight = ((double)words.size()*rand); // a random number in 0...words.size() - delta
		double weight = 0;
		
		String word = "";
		for (String w:words)
		{
			if (word.length() == 0)
			{
				weight = weight  + 1.0; // after the last word, weight = words.size()
				if (weight > topWeight) word = w;
			}
		}
		if (word.length() == 0) throw new Exception("Failed to return any word from leaf production in category " + cat.getName());
		return word;
	}


}
