package learnerPack;

import java.util.Vector;

import generatorPack.ExGenerator;
import modelPack.FSNode;
import modelPack.Ops;
import utilPack.Util;

/**
 * this class represents every occasion on which a learning example is used to learn a word
 *
 */

public class ExampleWord {
	
	// the learning example
	public LearningExample getLex() {return lex;}
	private LearningExample lex;
	
	// the word at the moment this ExampleWord was created
	private LearntWord lw;
	
	// distractor meaning used
	public boolean isDistractor() {return distractor;}
	private boolean distractor = false;
	
	// information content; must be computed at creation, as lw changes
	public int getInfo() {return info;}
	private int info;
	
	// if the example was used for repair of the word FS, rather than generalisation
	private boolean repair = false;
	public void setRepair(boolean repair) {this.repair = repair;}
	
	// productivity of the word
	private int[] prod = new int[2];
	
	// true if the meaning of the real word does not subsume the meaning of the learnt word
	public boolean isLoser() {return loser;}
	private boolean loser = false;
	
	// whether the learning example successfully updated the word
	public boolean succeeds() {return succeeds;}
	public void setSuccess(boolean succeeds) {this.succeeds = succeeds;}
	private boolean succeeds;
	
	// index of the unified FS within the learning example
	public int getFSNo() {return FSNo;}
	private int FSNo;
	
	private ExGenerator generator;
	
	//---------------------------------------------------------------------------
	//                            Constructor
	//---------------------------------------------------------------------------
	
	/**
	 * 
	 * @param lex
	 * @param lw
	 * @param FSNo
	 */
	public ExampleWord(LearningExample lex, LearntWord lw, int FSNo) throws Exception
	{
		this.lex = lex;
		this.lw = lw;
		this.FSNo = FSNo;
		succeeds = false;
		generator = lw.generator;
		
		// note that a distractor example meaning was used
		distractor = (lex.getbestMeaning() > 0);
		
		// set the productivity of the word
		prod = lw.getProductivity();
		
		// set the information count of the word
		FSNode wMean  = lw.getWord().getLastChild(); // meaning of the learnt word
		double ilearn = wMean.getInformation(generator.getInfoCount())*10.0;  // 10 times information content
		info = (int)ilearn;
		
		// note if the learnt word has lost any slots or nodes from the real word
		FSNode realWord = lw.realWord();
		if (realWord != null) // if there is a real word
		{
			FSNode realMeaning = realWord.getLastChild();
			if (!Ops.subsumes(realMeaning, wMean)) loser = true;
		}
	}
	
	//---------------------------------------------------------------------------
	//                            Methods
	//---------------------------------------------------------------------------
	
	/**
	 * 
	 * @return brief string summary, for writing out with the word
	 */
	public String stringForm() 
	{
		// example number
		String s = ("[Ex" + lex.getExNo());
		// if it was a distractor meaning
		if (distractor) s = s + ",dis";
		// repair
		if (repair) s = s + ",rep";
		// success or failure
		if (succeeds) s = s + ",pass"; else s = s + ",f";
		// productivity, if not [0,0]
		if ((prod[0]>0)||(prod[1]>0)) s = s + ",pr" + prod[0] + prod[1];
		// lost slots from real word
		if (loser) s = s + ",lo";
		// 10* meaning information content
		s = s + "," + info;
		
		return s + "]";	
	}
	
	/**
	 * true if the word occurs in the learning example isolated,
	 * with supported words on either side.
	 * This should give greater weight to a success.
	 * @return
	 */
	public boolean isolated()
	{
		boolean isol = false;
		Vector<String> wordSounds = lw.getWordSounds();		
		Vector<PartOfExample> wv = lex.getPartition();
		
		for (PartOfExample we: wv) if (we.isGap)
			if (Util.equalString(we.getWordSounds(), wordSounds)) isol = true;

		return isol;
	}
	
	/**
	 * 
	 * @return
	 */
	public int getExNo() {return lex.getExNo();}
	

}
