package learnerPack;

/**
 * Part of a Learning Example, which can either be a learnt word
 * or a sequence of sounds (a gap)
 */

import java.util.Vector;

import modelPack.FSNode;
import utilPack.Util;

public class PartOfExample {
	
	private Learner learner;
		
	Vector<String> wordSounds;
	public Vector<String> getWordSounds() {return wordSounds;}
	
	public LearntWord getLearntWord() {return lw;}
	LearntWord lw;
	
	int[] inputMeanings() 
	{
		if (isGap) return new int[2];
		return lw.getProductivity();
	}
	
	
	// true if this represents a gap rather than a supported word
	boolean isGap;
	
	//---------------------------------------------------------------------------
	//                                Constructors
	//---------------------------------------------------------------------------
	
	/**
	 *  constructor to represent uses of this example to refine candidate words,
	 *  or supported words in the partition of this example
	 * @param lw
	 */
	PartOfExample(LearntWord lw, Learner learner)
	{
		wordSounds = lw.getWordSounds();
		this.lw = lw;
		this.learner = learner;
		isGap = false;
	}
	
	/**
	 * constructor representing a gap (i.e not a supported word) in the sounds of this example
	 * @param sounds
	 */
	PartOfExample(Vector<String> sounds, Learner learner)
	{
		wordSounds = sounds;
		this.learner = learner;
		isGap = true;
	}

	
	//---------------------------------------------------------------------------
	//                               Methods
	//---------------------------------------------------------------------------
	
	/**
	 * 
	 * @return  all possible learnt words with the sounds of this word or sound sequence, 
	 * and with probability greater than a threshold;
	 * or if none can be found, return just this WordInExample (sounds, not a word)
	 */
	public Vector<PartOfExample> possibles()
	{
		double threshold = 0.1;
		Vector<PartOfExample> poss = new Vector<PartOfExample>();
		
		// get all LearntWords with the same sounds as this
		String spacedSound = FSNode.getSpacedName(wordSounds);		
		Vector<LearntWord> lwv = learner.getNamedWords(spacedSound);
				
		// filter words by probability
		for (LearntWord lw: lwv)
			if (lw.getProbability() > threshold) 
				poss.add(new PartOfExample(lw,learner));

		// make sure the list has at least one member - this one, if no other
		if (poss.size() == 0) poss.add(this);
		return poss;
	}
	
	
	boolean isSame(PartOfExample other)
	{
		boolean same = ((Util.includes(wordSounds, other.wordSounds)) && (Util.includes(other.wordSounds, wordSounds)));
		same = same & (inputMeanings()[0] == other.inputMeanings()[0]) & (inputMeanings()[1] == other.inputMeanings()[1]);
		return same;
	}
	
	
	// string form to write out
	String stringForm()
	{
		String res = "[";
		if (isGap) res = "<";

		for (String s: wordSounds) res = res + s + "-";
		
		if (isGap) res = res + "> ";
		else res = res + "] ";
		
		return res;
	}

			
}

