package modelPack;

import java.awt.Point;
import java.awt.Rectangle;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.swing.JFileChooser;
import javax.swing.filechooser.FileNameExtensionFilter;

// import javax.swing.JFileChooser;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import displayPack.EditLanguage;
import displayPack.Topic;
import displayPack.VFrame;
import utilPack.DataClass;
import utilPack.GenerateDataClass;
import utilPack.Util;

public class LanguageModel {
	
	private String location;
	
	private VFrame vFrame;
	public VFrame getFrame() {return vFrame;}
	
	// XML root elements of all data files
	private Hashtable<String,Element> fileRoots = new Hashtable<String, Element>();
	
	private Hashtable<String,NodeType> nodeTypes = new Hashtable<String,NodeType>();
	public Hashtable<String,NodeType> getNodeTypes() {return nodeTypes;}
	public NodeType getNodeType(String nodeName) {return nodeTypes.get(nodeName);}
	public void addNodeType(NodeType nt) {nodeTypes.put(nt.getName(), nt);}
	
	private Hashtable<String,Slot> slots = new Hashtable<String,Slot>();
	public Hashtable<String,Slot> getSlots() {return slots;}
	public Slot getSlot(String slotName) {return slots.get(slotName);}
	
	private Hashtable<String,FSNode> words = new Hashtable<String,FSNode> ();
	public Hashtable<String,FSNode>  getWords() {return words;}
	
	private Hashtable<String,Vector<FSNode>> wordSenses = new Hashtable<String,Vector<FSNode>>();
	public Vector<FSNode> getSenses(String wordName) {return wordSenses.get(wordName);}
	
	private Hashtable<String,FSNode> sentences = new Hashtable<String,FSNode>();
	public Hashtable<String,FSNode> getSentences() {return sentences;}
	
	// a list of sentence FS, in the order sent to the web client
	private Vector<FSNode> sentenceList = new Vector<FSNode>();
	public FSNode getSentence(int i) {return sentenceList.get(i);}
	
	private Hashtable<String,LearningStage> learningStageTable = new Hashtable<String,LearningStage>();
	public Hashtable<String,LearningStage> getLearningStageTable() {return learningStageTable;}
	public LearningStage getLearningStage(String name) {return learningStageTable.get(name);}
	
	private Vector<LearningStage> learningStages = new Vector<LearningStage>();
	public Vector<LearningStage> getLearningStages() {return learningStages;}
	
	// changing the Learning menu after a cycle of automatic learning
	public void clearLearningStages() 
	{
		learningStages = new Vector<LearningStage>();
		learningStageTable = new Hashtable<String,LearningStage>();
	}
	public void addLearningStage(LearningStage stage, String stageName) 
	{
		learningStages.add(stage);
		learningStageTable.put(stageName, stage);
	}
	
	private Vector<Topic> topics = new Vector<Topic>();
	public Vector<Topic> getTopics() {return topics;}
	
	private Hashtable<String,Topic> leafTopics = new Hashtable<String,Topic>();
	public Topic getLeafTopic(String id) {return leafTopics.get(id);}
	
	// the user guide topic which is shown at startup
	private String[] startTopicName = {"How Language Works"};	
	private String startTopicId = "";
	public String getStartTopicId() {return startTopicId;}
	
	// non-repeating ids for topics
	private int idIndex = 0;
	public String getId()
	{
		String id = "k" + idIndex;
		idIndex++;
		return id;
	}
	
	// the class that can hold all model data, if it is not to be read from files
	private DataClass dataClass = new DataClass();
	public DataClass getDataClass() {return dataClass;}
	
	//-------------------------------------------------------------------------------
	//                               Constructor
	//-------------------------------------------------------------------------------
	
	
	/**
	 * 
	 * @param viewFrame
	 * @param location file location of the model file
	 */
	public LanguageModel(VFrame viewFrame, String location)
	{
		this.location = location;
		this.vFrame = viewFrame;
	}
	
	//-------------------------------------------------------------------------------
	//                      Reading the language model
	//-------------------------------------------------------------------------------
	
	public static int MODELFILE = 0;
	public static int WORDFILE = 1;
	public static int SENTENCEFILE = 2;
	public static int LEARNINGFILE = 3;
	public static int GUIDEFILE = 4;
	
	public static String DATACLASSFILENAME = "Generated/DataClass.java";
	public static String LEARNTDATACLASSFILENAME = "Generated/LearntDataClass.java";
	public static String REALWORDCOMPAREFILENAME = "LearningResults/realWords.csv";
	public static String LEARNTWORDMEANINGFILENAME = "LearningResults/learntMeanings.txt";
	public static String PROGRESSSUMMARYFILENAME = "LearningResults/progress.csv";
	public static String SUPPORTHISTORYFILENAME = "LearningResults/support.csv";
	public static String EXAMPLELOGFILENAME = "LearningResults/exampleLog.txt";
	public static String SUBSUMPTIONTREEFILENAME = "LearningResults/subsumptionTrees.txt";
	
	public String dataClassFilePath = "";	
	public String learntDataClassFilePath = "";
	public String realWordComparePath = "";
	public String learntWordMeaningPath = "";
	public String progressSummaryPath = "";
	public String supportHistoryPath = "";
	public String exampleLogPath = "";
	public String subsumptionTreePath = "";
	
	// expected types of XML data file
	private String[] types = {"model","words","sentences","learning","guide","generator"};
	private Hashtable<String,String> fileNames = new Hashtable<String, String>();
	private Hashtable<String,String> filePaths = new Hashtable<String, String>();
	
	/**
	 * read the model files and check they are all present and OK as XML
	 * @param fromDataClass if true, read from the data class; if false read from files
	 * @throws Exception
	 */
	public void readModel(boolean fromDataClass) throws Exception
	{
		if (fromDataClass)
		{			
			readModelFromDataClass();
			return;
		}
		
		fileRoots = new Hashtable<String, Element>();

		Vector<String> fileLines = new Vector<String>();
				
		// try to find the text file of data file names at the hard-coded location
		try
		{
			fileLines = Util.textLines(location);
		}
		// if you do not find it, ask the user where to find it
		catch (Exception ex)
		{
			location = getModelFileLocation();
			if (location != null) fileLines = Util.textLines(location);
		}
		
		// if the user cancelled
		if (fileLines.size() == 0) return;
		
		// read the text file defining types and names of model XML files
		for (String line: fileLines)
		{
			StringTokenizer st = new StringTokenizer(line," ");
			if (st.countTokens() == 2)
			{
				String type = st.nextToken();
				String fileName = st.nextToken(); // just the file name; no path
				fileNames.put(type, fileName); 
			}
		}
		
		// check that all the required file types were defined in the text file
		for (String type: types) 
			if (fileNames.get(type) == null) 
				throw new Exception("No data file of type '" + type + "' defined in text file at " + location);
		
		// find all data files in the same folder as the text file
		File textFile = new File(location);		
		File folder = textFile.getParentFile();

		String folderPath = folder.getPath();
		dataClassFilePath = folderPath + "/" + DATACLASSFILENAME; // for saving the DataClass source code
		learntDataClassFilePath = folderPath + "/" + LEARNTDATACLASSFILENAME; // for saving the LearntDataClass source code
		realWordComparePath = folderPath + "/" + REALWORDCOMPAREFILENAME; // for csv comparison of slots in learnt words with real words
		learntWordMeaningPath = folderPath + "/" + LEARNTWORDMEANINGFILENAME; // for text output of FS of learnt words
		progressSummaryPath = folderPath + "/" + PROGRESSSUMMARYFILENAME;// for small tabular summary of learning progress
		supportHistoryPath = folderPath + "/" + SUPPORTHISTORYFILENAME;// for small tabular summary of learning progress
		exampleLogPath = folderPath + "/" + EXAMPLELOGFILENAME; // log of all learning examples
		subsumptionTreePath = folderPath + "/" + SUBSUMPTIONTREEFILENAME; // subsumption trees of learning examples for each word

		File[] siblings = folder.listFiles();
		
		// find all the XML root elements
		for (Enumeration<String> en = fileNames.keys(); en.hasMoreElements();)
		{
			String type = en.nextElement(); // 'model', 'words' etc.
			String  fileName = fileNames.get(type);
			
			InputStream stream = null;
			for (File sibling: siblings)
			{
				if (sibling.getName().equals(fileName)) 
				{
					// keep the location for later saving
					String filePath = folderPath + "/" + fileName;
					filePaths.put(type, filePath);
					try
					{
						stream  = new FileInputStream(sibling);

						Element root = Util.getRootElement(stream);
						fileRoots.put(type, root);
					}
					catch (Exception ex) {throw new Exception("Failed to read XML file " + fileName + "; " + ex.getMessage());}
				}
			}
			if (stream == null) throw new Exception("Could not find file " + fileName);						
		}
		
		// check that all the required file types were supplied
		for (String type: types) 
			if (fileRoots.get(type) == null) throw new Exception("No data file of type '" + type + "'");
		
		// read the files into the model
		readModelParts();
		
		
	}
	
	/**
	 * if the supplied file location does not have a file,
	 * use a file dialogue to get the location of the text file defining data file names
	 * @return
	 * @throws Exception
	 */
	private String getModelFileLocation() throws Exception
	{
		String modelFileLocation = null;
		JFileChooser modelFileChooser = new JFileChooser();
		modelFileChooser.setDialogTitle("Choose Language Model Text File");
		FileNameExtensionFilter filter = new FileNameExtensionFilter("txt","txt");
		modelFileChooser.setFileFilter(filter);
		int returnVal = modelFileChooser.showOpenDialog(null);
		if (returnVal == JFileChooser.APPROVE_OPTION) 
			 modelFileLocation =  modelFileChooser.getSelectedFile().getPath();
			    
		return modelFileLocation;
	 }

	
	/**
	 * 
	 */
	private void readModelFromDataClass() throws Exception
	{
		fileRoots = new Hashtable<String, Element>();

		// find all the XML root elements
		for (String type:types)
		{
			int typeInt = Util.indexOf(types, type);			
			InputStream stream = DataClass.getDataStream(typeInt);
			Element root = Util.getRootElement(stream);
			fileRoots.put(type, root);
		}
				
		// read the files into the model
		readModelParts();		
	}

	
	
	private Vector<String> validationErrors = new Vector<String>();
	
	public void addError(String error) {validationErrors.add(error);}
	
	public void writeValidationErrors()
	{
		if (validationErrors.size() > 0)
		{
			String allErrors = "There are " + validationErrors.size()  + " validation errors in  the language definition:\n\n";
			for (String error: validationErrors)
			{
				allErrors = allErrors + error + "\n";
			}
			vFrame.getAct().explain(allErrors);		
		}
	}
	
	/**
	 * validate every feature structure in the language model
	 * @throws Exception
	 */
	public void validateLanguageModel() throws Exception
	{
		validationErrors = new Vector<String>();
		for (Enumeration<FSNode> en = words.elements();en.hasMoreElements();)
		{
			FSNode fs = en.nextElement();
			fs.validateSlots();
		}
		
		for (Enumeration<FSNode> en = sentences.elements();en.hasMoreElements();)
		{
			FSNode fs = en.nextElement();
			fs.validateSlots();
		}
		
		for (Enumeration<LearningStage> en = learningStageTable.elements();en.hasMoreElements();)
		{
			LearningStage ls = en.nextElement();
			for (WordToLearn wl: ls.getLearnWords())
				for (FSNode fs: wl.getExamples()) fs.validateSlots();
		}
	}
	
	/**
	 * 
	 * @throws Exception
	 */
	public void readModelParts() throws Exception
	{
		Element nodesEl = Util.firstNamedChild(fileRoots.get("model"), "nodes");
		readNodes(nodesEl);
		Util.message("Read " + nodeTypes.size() + " node types;");
		
		Element slotsEl = Util.firstNamedChild(fileRoots.get("model"), "slots");
		readSlots(slotsEl);
		Util.message("Read " + slots.size() + " slots;");
		
		Element wordsEl = Util.firstNamedChild(fileRoots.get("words"), "words");
		readWords(wordsEl);
		Util.message("Read " + words.size() + " words;");
		
		Element sentencesEl = Util.firstNamedChild(fileRoots.get("sentences"), "sentences");
		readSentences(sentencesEl);
		Util.message("Read " + sentences.size() + " sentences.");
		
		Element learningEl = Util.firstNamedChild(fileRoots.get("learning"), "learning");
		readLearnMenu(learningEl);

		Element guideEl = Util.firstNamedChild(fileRoots.get("guide"), "guide");
		readGuide(guideEl);
				
		// do this for the workbench, not for the web server
		if (vFrame != null)
		{
			vFrame.setModel(this);
			vFrame.fillMenus(this);
		}
		
	}
	
	/**
	 * 
	 * @return root node of XML generator model
	 * @throws Exception
	 */
	public Element readGenerator() throws Exception
	{
		Element genEl = fileRoots.get("generator");
		if (genEl == null) throw new Exception("No generator model data");
		Util.message("Read generator data");
		return genEl;		
	}
	
	//-------------------------------------------------------------------------------
	//                 Reading node types
	//-------------------------------------------------------------------------------
	
	private void readNodes(Element nodesEl) throws Exception
	{
		if (nodesEl == null) throw new Exception("No 'nodes' element in model file");
		
		for (Element nodeEl: Util.namedChildElements(nodesEl, "node"))
		{
			NodeType node = new NodeType(this,nodeEl);
			nodeTypes.put(node.getName(), node);
		}
	
	}
	//-------------------------------------------------------------------------------
	//                 Reading slots
	//-------------------------------------------------------------------------------
	
	private void readSlots(Element slotsEl) throws Exception
	{
		if (slotsEl == null) throw new Exception("No 'slots' element in model file");
		
		for (Element slotEl: Util.namedChildElements(slotsEl, "slot"))
		{
			Slot slot = new Slot(this,slotEl);
			slots.put(slot.getName(), slot);
		}
	}
	
	/**
	 * true if a slot has an open set of values
	 * @param slotName
	 * @return
	 */
	public boolean openSlot(String slotName)
	{
		if (slotName.equals("subsumes")) return true;
		if (slotName.equals("id")) return true;
		if (slotName.equals("order")) return true;
		if (slotName.equals("subsumes")) return true;
		if (slotName.equals("sound")) return true;
		
		return false;
	}
	
	
	//-------------------------------------------------------------------------------
	//                 Reading words
	//-------------------------------------------------------------------------------
	
	private void readWords(Element wordsEl) throws Exception
	{
		if (wordsEl == null) throw new Exception("No 'words' element in model file");
		
		for (Element wordEl: Util.namedChildElements(wordsEl, "word"))
		{
			String name = wordEl.getAttribute("name");
			Element scriptEl = Util.firstNamedChild(wordEl, "script");
			
			FSNode wordFS = new FSNode(this,name,scriptEl,0,null);
			if (words.get(wordFS.getName()) != null) throw new Exception("Duplicate word: " + wordFS.getName());
			
			// add the word and save its sense
			addFS(wordFS, EditLanguage.WORD);
		}
	}
	
	
	//-------------------------------------------------------------------------------
	//                 Reading sentences
	//-------------------------------------------------------------------------------
	
	private void readSentences(Element sentEl) throws Exception
	{
		if (sentEl == null) throw new Exception("No 'sentences' element in model file");
		
		for (Element sentenceEl: Util.namedChildElements(sentEl, "sentence"))
		{
			String name = sentenceEl.getAttribute("name");
			Element scriptEl = Util.firstNamedChild(sentenceEl, "script");
			
			FSNode sentenceFS = new FSNode(this,name,scriptEl,0,null);
			if (sentences.get(sentenceFS.getName()) != null) throw new Exception("Duplicate sentence: " + sentenceFS.getName());
			sentences.put(sentenceFS.getName(),sentenceFS);
		}
		
	}
	
	//-------------------------------------------------------------------------------
	//                           Learning Menu
	//-------------------------------------------------------------------------------
	
	public void readLearnMenu(Element learningEl) throws Exception
	{
		Vector<Element> stageEls = Util.childElements(learningEl);
		
		for (Element stageEl: stageEls)
		{
			// null = there is no automatic learner at this point
			LearningStage stage = new LearningStage(this,null, stageEl);
			learningStageTable.put(stage.getName(), stage);
			learningStages.add(stage);
		}
		
	}
	
	
	
	//-------------------------------------------------------------------------------
	//                 Reading the guide
	//-------------------------------------------------------------------------------
	
	private void readGuide(Element guideEl)
	{
		Vector<Element> topicEls = Util.childElements(guideEl);
		
		for (Element  topicEl: topicEls) 
		{
			Topic top = new Topic(this,topicEl,null);
			top.storeTopicIds(leafTopics);
			topics.add(top);			
		}
		
		startTopicId = getLeafTopicId(topics,0);
		
	}
	
	/**
	 * 
	 * @param tops
	 * @param depth
	 * @return
	 */
	private String getLeafTopicId(Vector<Topic> tops, int depth)
	{
		for (Topic top: tops)
			if (top.getName().equals(startTopicName[depth]))
			{
				if (depth == startTopicName.length - 1) return top.getId();
				else return getLeafTopicId(top.getSubTopics(),depth + 1);				
			}
		
		// failure
		return "";
	}
	
	
	//-------------------------------------------------------------------------------
	//                            Editing and saving
	//-------------------------------------------------------------------------------
	
	
	/**
	 * add a newly created FS to a collection, so it can be saved
	 * @param fs
	 * @param type
	 */
	public void addFS(FSNode fs, int type) throws Exception
	{
		String name = fs.getName();
		if (type == EditLanguage.WORD) 
		{
			// make sure the word has the attribute result= 'true' on its final scene, if it has one
			Vector<FSNode> scenes = fs.getChildren();
			if (scenes.size() > 0) checkResultScene(fs);
						
			// save the word
			words.put(name, fs);
			
			// save the sense of the word
			StringTokenizer st = new StringTokenizer(name,"()");
			String stem = st.nextToken();
			Vector<FSNode> senses = getSenses(stem);
			if (senses == null) senses = new Vector<FSNode>();
			senses.add(fs);
			wordSenses.put(stem, senses);
		}
		
		//save a sentence
		else if (type == EditLanguage.SENTENCE) sentences.put(name, fs);
	}
	
	/**
	 * check that the last scene of a word FS is a result scene
	 * @param fs
	 */
	public void checkResultScene(FSNode fs) throws Exception
	{
		// make sure the word has the attribute result= 'true' on its final scene, if it has one
		Vector<FSNode> scenes = fs.getChildren();
		if (scenes.size() > 0)
		{
			FSNode lastScene = scenes.get(scenes.size() - 1);
			if (!(lastScene.getType().equals("scene"))) throw new Exception("Word FS last child node is not a scene");
			lastScene.setIsResult(true);
		}
	}
	
	/**
	 * 
	 * @throws Exception
	 */
	private void saveDataClass() throws Exception
	{
		boolean toFile = false;
		boolean forClient = false;

		// save the XML data files as strings
		String guideXML = saveGuideFile(toFile);
		String modelXML = saveModelFile(toFile);
		String wordXML = saveFile(EditLanguage.WORD,toFile,forClient);
		String sentenceXML = saveFile(EditLanguage.SENTENCE,toFile,forClient);
		String learnXML = saveFile(EditLanguage.LEARNINGEXAMPLE,toFile,forClient);
		
		// save XML form of menus as a string in the data class - to send to the web client
		Document menuDoc = Util.makeOutDoc();
		Element menuRoot = vFrame.getMenus().xmlForm(menuDoc);
		menuDoc.appendChild(menuRoot);
		String menuXML = dataFromXMLDocument(menuDoc);
		
		// pass the XML strings to the generator class, and generate the lines of DataClass.java
		GenerateDataClass generator = new GenerateDataClass(guideXML, modelXML, wordXML, sentenceXML,learnXML,menuXML);
		Vector<String> lines = generator.generateDataLines();
		
		// save DataClass.java
		Util.writeTextFile(dataClassFilePath, lines);
	}
	
	/**
	 * when save file methods are given toFile = false, they call this method
	 * @param doc an XML document
	 * @return the document converted to an unformatted string,
	 * with escape characters for a Java compiler
	 * @throws Exception
	 */
	public static String dataFromXMLDocument(Document doc) throws Exception
	{
		ByteArrayOutputStream stream = new ByteArrayOutputStream();
		Util.writeToStream(doc, stream, false); // write unformatted XML
		return addEscapes(stream.toString());
	}
	
	/**
	 * add escape characters '\' before any double quote '"',
	 * so it will be accepted as a single string by a Java compiler.
	 * Fails if the string starts or ends with '"' 
	 * - I could handle it; but it will not happen for XML strings
	 * @param raw
	 * @return
	 */
	public static String addEscapes(String raw)
	{
		// first check there are no reverse slashes introduced by edits; if so, remove them
		String noSlash = "";
		StringTokenizer st = new StringTokenizer(raw,"\\");
		while (st.hasMoreTokens())
		{
			String part = st.nextToken();
			noSlash = noSlash + part;
		}
							
		// now add reverse slashes before double quotes
		String result = "";
		st = new StringTokenizer(noSlash,"\"",true);
		while (st.hasMoreTokens())
		{
			String part = st.nextToken();
			
			// replace any '"' by '\"'; and replace '""' by '\"\"'
			if (part.equals("\"")) result = result + "\\" + "\"";
			else result = result + part;
		}
		return result;
	}


	
	/**
	 * 
	 * @throws Exception
	 */
	public void saveModelAndGuide() throws Exception
	{
		saveGuideFile(true);
		saveModelFile(true);
		
		// save the latest data class every time the model is saved
		saveDataClass();
	}
	
	/**
	 * 
	 * @param toFile
	 * @return
	 * @throws Exception
	 */
	public String saveGuideFile(boolean toFile) throws Exception
	{
		String result = null;
		String filePath = filePaths.get(types[GUIDEFILE]);
		if (toFile) Util.message("Saving model definition file at path " + filePath);

		if (filePath != null)
		{
			Document saveDoc = Util.makeOutDoc();
			
			// top element is always 'model'
			Element saveRoot  = Util.newElement(saveDoc, "model");
			saveDoc.appendChild(saveRoot);

			Element guideRoot = Util.newElement(saveDoc, "guide");
			saveRoot.appendChild(guideRoot);
			
			for (Topic top: topics) 
				guideRoot.appendChild(top.xmlForm(saveDoc));
			
			if (toFile) 
			{
				// formatted XML output  to a file
				Util.writeOutput(saveDoc, filePath, true);
			}
			else
			{
				result = dataFromXMLDocument(saveDoc);
			}
		}
		return result;
	}
	
	/**
	 * 
	 * @param toFile
	 * @return
	 * @throws Exception
	 */
	public String saveModelFile(boolean toFile) throws Exception
	{
		String result  = null;
		String filePath = filePaths.get(types[MODELFILE]);
		if (toFile) Util.message("Saving model definition file at path " + filePath);

		if (filePath != null)
		{
			Document saveDoc = Util.makeOutDoc();
			
			// top element is always 'model'
			Element saveRoot  = Util.newElement(saveDoc, "model");
			saveDoc.appendChild(saveRoot);
			
			// save all node types
			Element nodeRoot = Util.newElement(saveDoc, "nodes");
			saveRoot.appendChild(nodeRoot);
			for (Enumeration<NodeType> en = nodeTypes.elements(); en.hasMoreElements();)
			{
				NodeType nt = en.nextElement();
				nodeRoot.appendChild(nt.xmlForm(saveDoc));
			}
			
			// save all slots
			Element slotRoot = Util.newElement(saveDoc, "slots");
			saveRoot.appendChild(slotRoot);
			for (Enumeration<Slot> en = slots.elements(); en.hasMoreElements();)
			{
				Slot slot = en.nextElement();
				slotRoot.appendChild(slot.xmlForm(saveDoc));
			}
			
			if (toFile) 
			{
				// formatted XML output  to a file
				Util.writeOutput(saveDoc, filePath, true);
			}
			else
			{
				result = dataFromXMLDocument(saveDoc);
			}
		}
		return result;
	}
	
	/**
	 * save all three language files
	 * @throws Exception
	 */
	public void saveLanguageFiles() throws Exception
	{
		boolean toFile = true;
		boolean forClient = false;

		saveFile(EditLanguage.WORD,toFile,forClient);
		saveFile(EditLanguage.SENTENCE,toFile,forClient);
		saveFile(EditLanguage.LEARNINGEXAMPLE,toFile,forClient);

		// save the latest data class every time the language files are saved
		saveDataClass();
	}
	
	
	/**
	 * save a language file of given type in its current location
	 * @param type
	 * @param toFile
	 * - if true, save to a file and return null
	 * - if false, return a string of unformatted XML
	 * 
	 */
	public String saveFile(int type, boolean toFile,boolean forClient) throws Exception
	{
		String result = null;
		Hashtable<String,FSNode>  saveable = null;
		String fileType = "";
		String wrapName="";
		if (type == WORDFILE) 
		{
			fileType  = "words";
			wrapName = "word";
			saveable = words;
		}
		else if (type == SENTENCEFILE) 
		{
			fileType  = "sentences";
			wrapName = "sentence";
			saveable = sentences;
		}
		else if (type == LEARNINGFILE) 
		{
			fileType  = "learning";
		}
		else throw new Exception("Cannot save language model file of type " + type);
		
		String filePath = filePaths.get(fileType);
		Util.message("Saving file of type '" + fileType + "' at path " + filePath);
		
		if (filePath != null)
		{
			Document saveDoc = Util.makeOutDoc();
			
			// top element is always 'model'
			Element saveRoot  = Util.newElement(saveDoc, "model");
			saveDoc.appendChild(saveRoot);
			
			// next inner element is 'words' or 'sentences' or 'learning'
			Element outerEl = Util.newElement(saveDoc, fileType);
			saveRoot.appendChild(outerEl);
			
			if ((type == WORDFILE)|(type == SENTENCEFILE))
			{
				addWords(saveDoc,outerEl,saveable,wrapName,false);
			}
			
			else if (type == LEARNINGFILE) 
			{
				for (LearningStage stage:learningStages)
					outerEl.appendChild(stage.xmlForm(saveDoc,forClient));
			}
			
			if (toFile) 
			{
				// formatted XML output  to a file
				Util.writeOutput(saveDoc, filePath, true);
			}
			else
			{
				result = dataFromXMLDocument(saveDoc);
			}
			
		}
		return result;
	}
	
	/**
	 * 
	 * @param saveDoc
	 * @param outerEl
	 * @param saveable
	 * @param wrapName
	 */
	private Vector<FSNode> addWords(Document saveDoc,Element outerEl,Hashtable<String, FSNode>saveable,String wrapName, boolean forClient) throws Exception
	{
		Vector<FSNode> words = new Vector<FSNode>();
		for (Enumeration<FSNode> en = saveable.elements();en.hasMoreElements();)
		{
			FSNode toSave = en.nextElement();
			words.add(toSave);
			
			if (forClient)
			{
				toSave.setRow(-1);
				toSave.setColumn(0);
				toSave.setAllTruePositions();
			}
			
			// each FS is wrapped in an element 'word' or 'sentence'
			Element wrapEl = Util.newElement(saveDoc, wrapName);
			wrapEl.setAttribute("name", toSave.getName());
			outerEl.appendChild(wrapEl);
			
			// finally we get the XML for the feature structure
			wrapEl.appendChild(toSave.xmlForm(saveDoc,forClient));
		}
		return words;
	}
	
	//-------------------------------------------------------------------------------
	//           Sending an XML version of the complete model to the web client
	//-------------------------------------------------------------------------------
	
	/**
	 * 
	 * @param doc
	 * @return
	 */
	public Element clientModel(Document doc) throws Exception
	{
		// Util.message("Starting to compose model");
		Element root  = doc.createElement("model");

		// Words
		Element wordsEl = Util.newElement(doc, "words");
		root.appendChild(wordsEl);
		addWords(doc,wordsEl,words,"word",true);
		
		//  Sentences
		Element sentenceEl = Util.newElement(doc, "sentences");
		root.appendChild(sentenceEl);
		// save a list of sentences, in the order sent to the client
		sentenceList = addWords(doc,sentenceEl,sentences,"sentence",true);

		// Learning
		Element learnEl = Util.newElement(doc, "learning");
		root.appendChild(learnEl);
		boolean forClient  = true;
		for (LearningStage stage:learningStages)
			learnEl.appendChild(stage.xmlForm(doc,forClient));
		
		// node types
		Element nodeRoot = Util.newElement(doc, "nodes");
		root.appendChild(nodeRoot);
		for (Enumeration<NodeType> en = nodeTypes.elements(); en.hasMoreElements();)
		{
			NodeType nt = en.nextElement();
			nodeRoot.appendChild(nt.xmlForm(doc));
		}
		
		// slots
		Element slotRoot = Util.newElement(doc, "slots");
		root.appendChild(slotRoot);
		for (Enumeration<Slot> en = slots.elements(); en.hasMoreElements();)
		{
			Slot slot = en.nextElement();
			slotRoot.appendChild(slot.xmlForm(doc));
		}
		
		// User Guide
		Element guideRoot = Util.newElement(doc, "guide");
		root.appendChild(guideRoot);
		
		for (Topic top: topics) 
			guideRoot.appendChild(top.xmlForm(doc));
		
		// Menus - get XML from DataClass, as vFrame is null
		int menuIndex = 5;
		InputStream stream = DataClass.getDataStream(menuIndex);
		Element mRoot = Util.getRootElement(stream);
		Element menuRoot = (Element)doc.importNode(mRoot, true);
		root.appendChild(menuRoot);
		
		// Util.message("Server made up model with " + wordCount + " words and " + sentenceCount + " sentences.");
		return root;
	}
	
	
	
	//-------------------------------------------------------------------------------
	//                 Drawn FS nodes or SlotValues selected by mouse click
	//-------------------------------------------------------------------------------
	
	// clicked nodes on Feature Structures - built up when Feature Structures are drawn
	
	private Hashtable<Rectangle,FSNode> nodeLocations = new Hashtable<Rectangle,FSNode>();
	
	public void clearNodeLocations() {nodeLocations = new Hashtable<Rectangle,FSNode>();}
	
	public void addNodeLocation(Rectangle loc, FSNode fs) {nodeLocations.put(loc, fs);};
	
	
	// clicked slot values - built up when a SlotValue is drawn
	
	private Hashtable<Rectangle,SlotValue> valueLocations = new Hashtable<Rectangle,SlotValue>();
	
	public void clearValueLocations() {valueLocations = new Hashtable<Rectangle,SlotValue>();}
	
	public void addValueLocation(Rectangle loc, SlotValue sv) {valueLocations.put(loc, sv);};
	

	// clicked [+-] boxes - built up when a SlotValue is drawn
	
	private Hashtable<Rectangle,SlotValue> pmBoxLocations = new Hashtable<Rectangle,SlotValue>();
	
	public void clearpmBoxLocations() {pmBoxLocations = new Hashtable<Rectangle,SlotValue>();}
	
	public void addpmBoxLocation(Rectangle loc, SlotValue sv) {pmBoxLocations.put(loc, sv);};
		
	/**
	 * 
	 * @param x
	 * @param y
	 * @return
	 */
	public FSNode getClickedNode(int x, int y)
	{
		FSNode res = null;
		
		for (Enumeration<Rectangle> en = nodeLocations.keys();en.hasMoreElements();)
		{
			Rectangle loc = en.nextElement();
			if (loc.contains(new Point(x,y))) res = nodeLocations.get(loc);
		}		
		return res;
	}
	
	
/**
 * 
 * @param x
 * @param y
 * @return
 */
public SlotValue getClickedValue(int x, int y)
{
	SlotValue res = null;
	
	for (Enumeration<Rectangle> en = valueLocations.keys();en.hasMoreElements();)
	{
		Rectangle loc = en.nextElement();
		if (loc.contains(new Point(x,y))) res = valueLocations.get(loc);
	}		
	return res;
}

/**
 * 
 * @param x
 * @param y
 * @return
 */
public SlotValue getClickedpmBox(int x, int y)
{
	SlotValue res = null;
	
	for (Enumeration<Rectangle> en = pmBoxLocations.keys();en.hasMoreElements();)
	{
		Rectangle loc = en.nextElement();
		if (loc.contains(new Point(x,y))) res = pmBoxLocations.get(loc);
	}		
	return res;
}
	
	//-------------------------------------------------------------------------------
	//                 Miscellania
	//-------------------------------------------------------------------------------
//-------------------------------------------------------------------------------
//                                     Tests
//-------------------------------------------------------------------------------

private void testUnification() throws Exception
{
	Util.message("\nUnification test");
			
	/* FeatureStructure boy = words.get(0);
	FeatureStructure runs = words.get(4);
	FeatureStructure sent = sentences.get(0);

	
	FeatureStructure first = Ops.unify(boy,sent, Ops.UNDERSTAND);
	first.write("");
	
	FeatureStructure second = Ops.unify(runs,first, Ops.UNDERSTAND);
	second.write("");
	*/
			
}




	

//-------------------------------------------------------------------------------
//                             Choosing file locations  (not used??)
//-------------------------------------------------------------------------------

/**
 * 
 * @param title
 * @return
private String chooseNewFile(String title)
{
	String location = null;
	JFileChooser modelFileChooser = new JFileChooser();
	modelFileChooser.setDialogTitle(title);
	modelFileChooser.setFileFilter(filter);
	
	int returnVal = modelFileChooser.showOpenDialog(null);
	if (returnVal == JFileChooser.APPROVE_OPTION) 
		{location =  modelFileChooser.getSelectedFile().getPath();}
	return location;
}
*/



}
