package modelPack;

import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import learnerPack.Learner;
import utilPack.Util;

public class LearningStage {
	
	private LanguageModel model;
	public LanguageModel getModel() {return model;}
	
	private String stageName;
	public String getName() {return stageName;}
	
	private int cycleNo = 0;
	public int getCycleNo() {return cycleNo;}
	
	// automatic learner
	private Learner learner;
	public Learner getLearner() {return learner;}

	
	private Vector<WordToLearn> learnTable = new Vector<WordToLearn>();
	public Vector<WordToLearn> getLearnWords() {return learnTable;}
	public WordToLearn getWordLearn(String wordName) 
	{
		WordToLearn res = null;
		for (WordToLearn wl:learnTable) 
			if (wl.getWordName().equals(wordName)) res = wl;
		return res;
	}
	
	public void addWord(WordToLearn wl) {learnTable.add(wl);}
	


	//--------------------------------------------------------------------------
	//                        Constructors
	//--------------------------------------------------------------------------
	
	
	/**
	 * from XML input
	 * @param model
	 * @param stageEl
	 * @throws Exception
	 */
	public LearningStage(LanguageModel model, Learner learner, Element stageEl) throws Exception
	{
		this.model = model;
		this.learner = learner;
		stageName = stageEl.getAttribute("name");
		String cycle = stageEl.getAttribute("cycleNo");
		if (cycle.length()  > 0) cycleNo = Integer.parseInt(cycle);

		Vector<Element> learnEls = Util.childElements(stageEl);
		
		for (Element learnEl: learnEls)
		{
			WordToLearn wLearn = new WordToLearn(this,learnEl,stageName);
			learnTable.add(wLearn);
		}	
	}
	
	/**
	 * after a cycle of automatic learning
	 * @param model
	 * @param stageName
	 * @throws Exception
	 */
	public LearningStage(LanguageModel model, String stageName, Learner learner, int cycleNo) throws Exception
	{
		this.model = model;
		this.stageName = stageName;
		this.cycleNo = cycleNo;
		this.learner = learner;
	}

	
	/**
	 * 
	 * @param doc
	 * @return
	 * @throws Exception
	 */
	public Element xmlForm(Document doc, boolean forClient) throws Exception
	{
		Element stageEl = Util.newElement(doc, "stage");
		stageEl.setAttribute("name", stageName);
		stageEl.setAttribute("cycleNo", Integer.valueOf(cycleNo).toString());
		
		for (WordToLearn wl: learnTable)
			stageEl.appendChild(wl.xmlForm(doc,forClient));
		
		return stageEl;
	}


}
