package modelPack;

import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import utilPack.Util;

public class NodeType {
	
	private String name;
	public String getName() {return name;}
	
	private String description = "";
	public String getDescription() {return description;}
	public void setDescription(String desc) {description = desc;}
	
	private Vector<String> childNodeNames = new Vector<String>();
	public Vector<String> getChildNodeNames() {return childNodeNames;}
	public boolean isAllowedChild(String type) {return Util.inVector(type, childNodeNames);}
	
	/**
	 * 
	 * @return allowed slots for this node type
	 */
	public Vector<Slot> allowedSlots()
	{
		Vector<Slot> allowed = new Vector<Slot>();
		
		for (Enumeration<Slot> en = model.getSlots().elements();en.hasMoreElements();)
		{
			Slot slot = en.nextElement();
			if (slot.isAllowedNodeType(name)) allowed.add(slot);
		}
		
		return allowed;
	}
	
	private LanguageModel model;
	
	/**
	 * true if the child node type is allowed
	 * @param nodeName
	 * @return
	 */
	public boolean isValidChild(String nodeName)
	{
		boolean valid = false;
		for (String name: childNodeNames)
			if (name.equals(nodeName)) valid = true;
		return valid;
	}
	
	//--------------------------------------------------------------------------
	//                        Constructor
	//--------------------------------------------------------------------------
	
	/**
	 * constructor for making a new node type in the editor
	 * @param name
	 */
	public NodeType(LanguageModel model, String name)
	{
		this.model = model;
		this.name = name;
	}

	
	/**
	 * constructor from XML
	 * @param nodeEl
	 */
	public NodeType(LanguageModel model, Element nodeEl)
	{
		this.model = model;
		name = nodeEl.getAttribute("name");
		description = nodeEl.getAttribute("dsc");
		
		String children = nodeEl.getAttribute("children");
		setChildNodeTypes(children);
	}
	
	/**
	 * set the allowed child node types
	 * @param children
	 */
	public void setChildNodeTypes(String children)
	{
		childNodeNames = new Vector<String>();
		if (children.length() > 0)
		{
			StringTokenizer st = new StringTokenizer(children,", ");
			while (st.hasMoreTokens()) childNodeNames.add(st.nextToken());
		}
	}
	
	/**
	 * 
	 * @return
	 */
	public String getExplanation()
	{
		String explain = "Node type:  " + name;
		if (description.length() > 0) explain = explain + "\n\n" + description;

		explain = explain + "\n\nAllowed child node types: ";
		if (childNodeNames.size() == 0) explain = explain + " None";
		else for (String name: childNodeNames) {explain = explain + " " + name;}

		explain = explain + "\n\nAllowed slots: ";
		for (Enumeration<Slot> en = model.getSlots().elements(); en.hasMoreElements();)
		{
			Slot slot = en.nextElement();
			Vector<NodeType> types = slot.getAllowedTypes();
			for (NodeType type: types) 
				if (type.getName().equals(getName())) 
						explain = explain + slot.getName() + " ";
		}
		
		return explain;
	}
	
	//--------------------------------------------------------------------------
	//                         XML output
	//--------------------------------------------------------------------------
	
	/**
	 * 
	 * @param doc
	 * @return
	 * @throws Exception
	 */
	public Element xmlForm(Document doc) throws Exception
	{
		Element root = Util.newElement(doc, "node");
		root.setAttribute("name", name);
		root.setAttribute("dsc", description);
				
		String childNames = "";
		for (String cNode: childNodeNames) childNames = childNames + cNode + " ";
		root.setAttribute("children", childNames);
				
		return root;
	}

}
